# Fields used in PC data structures

PC_F_COST_CAT_ID         = 'costCatID'
PC_F_COST_INPUT_ID       = 'costInputID'
PC_F_AREA_ID             = 'areaID'
PC_F_AREA_TYPE_ID        = 'areaTypeID'
PC_F_COSTING_SECT_ID     = 'costingSectionID'
PC_F_COSTING_SUBSECT_ID  = 'costingSubsectionID'
PC_F_ACT_ID              = 'activityID'
PC_F_VALUE               = 'value'

# area data record (main object in PC; basically contains everything you need to know about PC for a given 
# programme area/delivery channel/package/national programme)

PC_AREA_TYPE_MST_ID = 'areaTypeMstID' #int; programme area, delivery channel, etc.
PC_AREA_ID          = 'areaID'        # string; represents programme area, delivery channel, package, or national programme; numeric mstID for defaults. GUID for customs.
PC_MAIN_ROW_COSTS   = 'mainRowCosts'  # list of float, row x year; PCDetailCostingArray on desktop

# cost category records; Groups to which user adds cost inputs.  

PC_COST_CAT_ID        = PC_F_COST_CAT_ID # string, numeric mstID for defaults (still cast as string). GUID for custom ones
PC_COST_CAT_NAME      = 'name'           # string;  Holds custom name or overridden default name.
PC_COST_CAT_STR_CONST = 'stringConstant' # string;  Defaults only. 
PC_COST_CAT_CUSTOM    = 'custom'         # boolean; user-entered = true; default = false.

# cost input records; Items belonging to cost categories that are used for activity line items.

PC_COST_INPUT_CAT_ID                    = PC_F_COST_CAT_ID              # string, numeric mstID for defaults (still cast as string). GUID for custom ones
PC_COST_INPUT_ID                        = PC_F_COST_INPUT_ID            # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_COST_INPUT_NAME                      = 'name'                        # string;  Holds custom name or overridden default name.
PC_COST_INPUT_STR_CONST                 = 'stringConstant'              # string;  Defaults only. 
PC_COST_INPUT_UNIT_OF_MEASURE           = 'unitOfMeasure'               # string; Activity cost per "x"
PC_COST_INPUT_UNIT_OF_MEASURE_STR_CONST = 'unitOfMeasureStringConstant' # string;  Defaults only. 
PC_COST_INPUT_UNIT_COST                 = 'unitCost'                    # float
PC_COST_INPUT_NOTE                      = 'note'                        # string
PC_COST_INPUT_SOURCE                    = 'source'                      # string
#PC_COST_INPUT_SOURCE_STR_CONST          = 'stringConstant'             # string;  Defaults only. 
PC_COST_INPUT_CUSTOM                    = 'custom'                      # boolean; user-entered = true; default = false.

# costing section records; Top-level groups for costing.

PC_COSTING_SECT_ID                                = PC_F_COSTING_SECT_ID  # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_COSTING_SECT_NAME                              = 'name'                # string;  Holds custom name or overridden default name.
PC_COSTING_SECT_STR_CONST                         = 'stringConstant'      # string;  Defaults only. 
PC_COSTING_SECT_CUSTOM                            = 'custom'              # boolean; user-entered = true; default = false.
PC_COSTING_SECT_SUBSECT_RECORDS                   = 'subsectionRecords'   # list; Subgroup dicts
PC_COSTING_SECT_AREA_TYPE_ID                      = PC_F_AREA_TYPE_ID     # integer; tells us whether it's a module, programme area, etc.
PC_COSTING_SECT_AREA_ID                           = PC_F_AREA_ID          # string; ID of module, programme area, delivery channel, national programme, etc.
PC_COSTING_SECT_ENTRY_MODE_CS                     = 'entryMode'
PC_COSTING_SECT_COSTS_CS                          = 'costs'              
# Policy and population level interventions (PPLI)
PC_COSTING_SECT_PRICE_ELASTICITY_PPLI             = 'priceElasticity'          # list of float (length 2, can be negative, calculated packs change as integer)
PC_COSTING_SECT_INCOME_ELASTICITY_PPLI            = 'incomeElasticity'         # list of float (length 2, calculated packs change as integer)
PC_COSTING_SECT_PRICE_INCOME_ONLY_PPLI            = 'priceIncomeOnly'          # list of integer (length 2, can be negative, percent, calculated packs change only)
PC_COSTING_SECT_NON_PRICE_MEASURES_PPLI           = 'nonPriceMeasures'         # list of integer (length 2, can be negative, percent)
PC_COSTING_SECT_ALL_MPOWER_PPLI                   = 'allMPOWER'                # list of integer (length 2, calculated packs change only)

PC_COSTING_SECT_TAX_ENTRY_MODE_PPLI               = 'taxEntryModeID'           # integer (radio group choice)
PC_COSTING_SECT_REAL_RETAIL_PRICE_PPLI            = 'realRetailPrice'          # list of integer (length 2, projection sometimes calculated)
PC_COSTING_SECT_EXCISE_SHARE_PPLI                 = 'exciseShare'              # list of integer (length 2, percent, projection sometimes calculated)
PC_COSTING_SECT_TAX_SHARE_PPLI                    = 'taxShare'                 # list of integer (length 2, percent, projection sometimes calculated)
PC_COSTING_SECT_REAL_EXCISE_PER_PACK_PPLI         = 'realExcisePerPack'        # list of integer (length 2, calculated)
PC_COSTING_SECT_REAL_TAX_PER_PACK_PPLI            = 'realTaxPerPack'           # list of integer (length 2, calculated)
PC_COSTING_SECT_REAL_INDUSTRY_PRICE_PPLI          = 'realIndustryPrice'        # list of integer (length 2, calculated)
PC_COSTING_SECT_TAX_ADJUST_CONSUMP_PPLI           = 'taxAdjustedConsumption'   # list of integer (length 2, projection calculated)
PC_COSTING_SECT_EXCISE_REVENUE_PPLI               = 'exciseRevenue'            # list of integer (length 2, projection calculated)
PC_COSTING_SECT_TAX_REVENUE_PPLI                  = 'taxRevenue'               # list of integer (length 2, projection calculated)
PC_COSTING_SECT_USAGE_RATE_PPLI                   = 'usageRate'                # list of float (length 2, percent, projection claculated as a float)
PC_COSTING_SECT_NUM_USERS                         = 'numUsers'                 # list of integer (length 2, calculated)

PC_COSTING_SECT_REAL_RETAIL_PRICE_PPLI            = 'realRetailPrice'          # list of integer (length 2)

# costing subsection records; Second-level groups for costing that belong to costing sections.

PC_COSTING_SUBSECT_ID         = PC_F_COSTING_SUBSECT_ID  # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_COSTING_SUBSECT_NAME       = 'name'                   # string;  Holds custom name or overridden default name.
PC_COSTING_SUBSECT_STR_CONST  = 'stringConstant'         # string;  Defaults only. 
PC_COSTING_SUBSECT_CUSTOM     = 'custom'                 # boolean; user-entered = true; default = false.
# Policy and population level interventions (PPLI)
PC_COSTING_SUBSECT_ADMIN_UNITS_PER_LEVEL_OPTION_PPLI          = 'numAdminUnitsPerLevelOption' # integer
PC_COSTING_SUBSECT_ADMIN_UNITS_STD_POP_SIZE_PPLI              = 'adminUnitsStdPopSize' # 1D list of integer; by health system admin unit
PC_COSTING_SUBSECT_ENTER_TOTAL_INPUTS_PER_ADMIN_REG_YEAR_PPLI = 'enterTotalInputsPerAdminRegYear' # boolean

PC_COSTING_SUBSECT_RAISE_TAXES_PRICE_PPLI            = 'raiseTaxesPrice'          # list of integer; by year
PC_COSTING_SUBSECT_RAISE_TAXES_EXCISE_SHARE_PPLI     = 'raiseTaxesExciseShare'    # list of integer; by year
PC_COSTING_SUBSECT_RAISE_TAXES_TAX_SHARE_PPLI        = 'raiseTaxesTaxShare'       # list of integer; by year
PC_COSTING_SUBSECT_PHASE_INTENSITY_VALUES_PPLI       = 'phaseIntensityValues'     # 2d list of integer; phase/intensity type x year

# activity records; Third-level groups for costing that belong to costing subsection records. Activity line items are added to these.

PC_ACT_ID                 = PC_F_ACT_ID             # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_ACT_NAME               = 'name'                  # string;  Holds custom name or overridden default name.
PC_ACT_STR_CONST          = 'stringConstant'        # string;  Defaults only. 
PC_ACT_AREA_TYPE_ID       = PC_F_AREA_TYPE_ID       # integer; tells us whether it's a module, programme area, etc.
PC_ACT_AREA_ID            = PC_F_AREA_ID            # string; ID of module, programme area, delivery channel, national programme, etc.
PC_ACT_COSTING_SECT_ID    = PC_F_COSTING_SECT_ID    # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_ACT_COSTING_SUBSECT_ID = PC_F_COSTING_SUBSECT_ID # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_ACT_CUSTOM             = 'custom'                # boolean; user-entered = true; default = false.

# activity line item (ALI) records; Items based on cost inputs that the user adds to activities.

PC_ALI_COST_INPUT_ID      = PC_F_COST_INPUT_ID   # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_ALI_ACT_ID             = PC_F_ACT_ID             # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_ALI_AREA_TYPE_ID       = PC_F_AREA_TYPE_ID       # integer; tells us whether it's a module, programme area, etc.
PC_ALI_AREA_ID            = PC_F_AREA_ID            # string; ID of module, programme area, delivery channel, national programme, etc.
PC_ALI_COSTING_SECT_ID    = PC_F_COSTING_SECT_ID    # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_ALI_COSTING_SUBSECT_ID = PC_F_COSTING_SUBSECT_ID # string; numeric mstID for defaults (still cast as string). GUID for custom ones
PC_ALI_NUM_UNITS          = 'numUnits'              # integer
PC_ALI_DURATION           = 'duration'              # integer
PC_ALI_FREQUENCY          = 'frequency'             # list of integer
# Policy and population level interventions (PPLI)
PC_ALI_QUANTITY_PER_ADMIN_REG_YEAR_PPLI = 'quantityPerAdminRegYear' # 2D list of float; health sys admin unit (dynamic) x year


import numpy as np
from copy import deepcopy


import SpectrumCommon.Const.GB.GBConst as gbc
import SpectrumCommon.Const.RP.RPConst as rpc
import SpectrumCommon.Const.RP.RPModVarFields as rpmvf
# import SpectrumCommon.Util.RP.RPUtilLink as rpul

# def RP_create_sector_rec(ID = int, on = bool, num_years = int, meta : bool = False):
def RP_create_sector_rec(ID = int, num_years = int, meta : bool = False):
    sector_rec = { 
        rpmvf.RP_SECTOR_ID           : ID,
        # rpmvf.RP_SECTOR_ON           : on, # Moved to GB since this info is shown in the Projection Manager which apparently can only use GB/PJ MVs
        rpmvf.RP_SECTOR_GOAL_RECORDS : RP_init_goal_records(ID, num_years, meta),
    }

    if ID == rpc.RP_EDUCATION_SECTOR_MST_ID:
        sector_rec[rpmvf.RP_SECTOR_EDU_PRIM_SCH_START_AGE] = 0
        sector_rec[rpmvf.RP_SECTOR_EDU_PRIM_SCH_END_AGE] = 0
        sector_rec[rpmvf.RP_SECTOR_EDU_SEC_SCH_START_AGE] = 0
        sector_rec[rpmvf.RP_SECTOR_EDU_SEC_SCH_END_AGE] = 0
        # sector_rec[rpmvf.RP_SECTOR_EDU_STUDENT_PER_TEACHER_PRIM] = 0.0
        sector_rec[rpmvf.RP_SECTOR_EDU_AVG_ANN_TEACHER_SAL_PRIM_GDP_RATIO] = 0
        # sector_rec[rpmvf.RP_SECTOR_EDU_AVG_ANN_TEACHER_SAL_SEC] = 0
        sector_rec[rpmvf.RP_SECTOR_EDU_PRIM_SPEND_PER_STUDENT_PERC_GDP] = 0
        sector_rec[rpmvf.RP_SECTOR_EDU_SEC_SPEND_PER_STUDENT_PERC_GDP] = 0

    elif ID == rpc.RP_HEALTH_SECTOR_MST_ID:
        sector_rec[rpmvf.RP_SECTOR_HLTH_AVG_UNIT_COST_SKILLED_DELIV] = 0.0
        sector_rec[rpmvf.RP_SECTOR_HLTH_AVG_UNIT_COST_ANC_4_PLUS_VISITS] = 0.0
        sector_rec[rpmvf.RP_SECTOR_HLTH_PERC_SKILLED_HLTH_PROF_DOCTORS] = 0
        sector_rec[rpmvf.RP_SECTOR_HLTH_PERC_SKILLED_HLTH_PROF_NURSES] = 0
        sector_rec[rpmvf.RP_SECTOR_HLTH_AVG_SALARY_DOCTORS] = 0.0
        sector_rec[rpmvf.RP_SECTOR_HLTH_AVG_SALARY_NURSES] = 0.0

    elif ID == rpc.RP_WASH_SECTOR_MST_ID:
        sector_rec[rpmvf.RP_SECTOR_WASH_CAP_COSTS_PER_HH_IMPR_WATER] = 0.0
        sector_rec[rpmvf.RP_SECTOR_WASH_ANN_COST_MENSTRUAL_HLTH_COMM] = 0.0
        sector_rec[rpmvf.RP_SECTOR_WASH_CAP_COSTS_PER_HH_IMPR_SANI] = 0.0

    elif ID == rpc.RP_FOOD_SECURITY_SECTOR_MST_ID:
        sector_rec[rpmvf.RP_SECTOR_FOOD_PRIM_STAPLE_CROP] = ''
        sector_rec[rpmvf.RP_SECTOR_FOOD_AGRICULT_LAND] = 0
        sector_rec[rpmvf.RP_SECTOR_FOOD_CROP_LAND] = 0
        sector_rec[rpmvf.RP_SECTOR_FOOD_STAPLE_CROP_PROD_BASELINE] = 0
        sector_rec[rpmvf.RP_SECTOR_FOOD_AREA_HARV_STAPLE_CROP] = 0
        sector_rec[rpmvf.RP_SECTOR_FOOD_TONNES_HARV_PERC_HECT] = 0.0
        sector_rec[rpmvf.RP_SECTOR_FOOD_ANN_KG_PER_YEAR_CONSUMP_STAPLE_CROP] = 0
        sector_rec[rpmvf.RP_SECTOR_FOOD_PROD_COSTS_STAPLE_CROP] = 0.0
        sector_rec[rpmvf.RP_SECTOR_FOOD_DAILY_COST_FOOD_AID] = 0.0

    elif ID == rpc.RP_URBANIZATION_SECTOR_MST_ID:
        sector_rec[rpmvf.RP_SECTOR_URB_PERC_POP_URB_BASELINE_YEAR] = 0
        sector_rec[rpmvf.RP_SECTOR_URB_PERC_POP_URB_ENDLINE_YEAR] = 0
        sector_rec[rpmvf.RP_SECTOR_URB_COST_PER_HH_MOVE_TO_HOUSING] = 0.0

    elif ID == rpc.RP_ENERGY_SECTOR_MST_ID:
        sector_rec[rpmvf.RP_SECTOR_ENERGY_KWH_PROD_PER_CAP_PER_YEAR] = 0.0
        sector_rec[rpmvf.RP_SECTOR_ENERGY_CAP_COSTS_PER_HH_CONNECT_GRID] = 0
        sector_rec[rpmvf.RP_SECTOR_ENERGY_CAP_COSTS_PER_HH_CLEAN_COOKING] = 0
        sector_rec[rpmvf.RP_SECTOR_ENERGY_COST_PER_KWH] = 0.0

    elif ID == rpc.RP_ECONOMIC_SECTOR_MST_ID:
        sector_rec[rpmvf.RP_SECTOR_ECO_GDP_PER_CAPITA] = 0.0
        # sector_rec[rpmvf.RP_SECTOR_ECO_BASELINE_GDP] = 0
        # sector_rec[rpmvf.RP_SECTOR_ECO_GDP_GROWTH_RATE] = 0.0
        sector_rec[rpmvf.RP_SECTOR_ECO_LABOR_PARTIC_RATE] = 0.0

    elif ID == rpc.RP_BUILD_YOUR_OWN_SECTOR_MST_ID:
        # sector_rec[rpmvf.RP_SECTOR_BYO_POP] = ''
        sector_rec[rpmvf.RP_SECTOR_BYO_INTERV] = ''
        sector_rec[rpmvf.RP_SECTOR_BYO_TARG_POP] = rpc.RP_TP_TOTAL_POP_MST_ID
        # sector_rec[rpmvf.RP_SECTOR_BYO_UNIT_OF_ANALYSIS_1] = ''
        sector_rec[rpmvf.RP_SECTOR_BYO_MULTIPLIER_1] = 0
        sector_rec[rpmvf.RP_SECTOR_BYO_UNIT_OF_ANALYSIS_2] = ''
        sector_rec[rpmvf.RP_SECTOR_BYO_UNIT_COST_2] = 0.0

    return sector_rec

def RP_create_goal_rec(sector_ID = int, goal_ID = int, num_years = int, meta : bool = False):
    goal_rec = {
        rpmvf.RP_GOAL_ON  : True,
        rpmvf.RP_GOAL_ID  : goal_ID,
    }

    RC_format = rpc.RP_SECTOR_GOAL_RC_FORMATS[sector_ID][goal_ID]

    if RC_format != rpc.RP_RC_NOT_APPLICABLE:
        if rpc.RP_SECTOR_DEFAULT_SUBGOALS[sector_ID]:
            if meta:
                goal_rec[rpmvf.RP_GOAL_DEFAULT_SUBGOAL_ID] = rpc.RP_SECTOR_DEFAULT_SUBGOALS[sector_ID][goal_ID]
            else:
                goal_rec[rpmvf.RP_GOAL_SELECTED_SUBGOAL_ID] = rpc.RP_SECTOR_DEFAULT_SUBGOALS[sector_ID][goal_ID]

    # Special case: Build your own sector has it's resulting change in the goal record, since it has no subgoals. All other sectors have at 
    # least one goal with a subgoal.
    if sector_ID == rpc.RP_BUILD_YOUR_OWN_SECTOR_MST_ID:
        goal_rec[rpmvf.RP_GOAL_RESULTING_CHANGE] = np.full((num_years), 0.0).tolist()
    else:
        goal_rec[rpmvf.RP_GOAL_SUBGOAL_RECORDS] = RP_init_subgoal_records(sector_ID, goal_ID, num_years, meta)

    return goal_rec

def RP_create_subgoal_rec(sector_ID = int, goal_ID = int, subgoal_ID = int, num_years = int):
    subgoal_rec = {
        rpmvf.RP_SUBGOAL_ID : subgoal_ID,
    }

    # Resulting change will take one of three forms depending on the goal: sex x time, time, or non-existant
    RC_format = rpc.RP_SECTOR_GOAL_RC_FORMATS[sector_ID][goal_ID]

    if RC_format == rpc.RP_RC_BY_SEX_YEAR:
        subgoal_rec[rpmvf.RP_SUBGOAL_RESULTING_CHANGE] = np.full((gbc.GB_NumSexes, num_years), 0.0).tolist()

    elif RC_format == rpc.RP_RC_BY_YEAR:
        subgoal_rec[rpmvf.RP_SUBGOAL_RESULTING_CHANGE] = np.full((num_years), 0.0).tolist()

    return subgoal_rec

def RP_create_climate_params_rec():
    return {
        rpmvf.RP_CLIM_BASELINE_MEAN_MED_TEMP              : 0.0,
        rpmvf.RP_CLIM_PERC_CHANGE_GDP_1_DEG_WARM          : 0.0,
        rpmvf.RP_CLIM_PERC_CHANGE_ENERGY_NEEDS_1_DEG_WARM : 0.0,
        rpmvf.RP_CLIM_PERC_CHANGE_STAPLE_CROP_1_DEG_WARM  : 0.0,
        rpmvf.RP_CLIM_PERC_CHANGE_STAPLE_CROP_2_DEG_WARM  : 0.0,
        rpmvf.RP_CLIM_PERC_CHANGE_STAPLE_CROP_3_DEG_WARM  : 0.0,
    }

# def RP_init_sector_records(sector_on = list, num_years = int, meta : bool = False):
def RP_init_sector_records(num_years = int, meta : bool = False):
    sectors = []

    for sector_ID in rpc.RP_SECTOR_MST_IDS:
        if (not meta) or (sector_ID != rpc.RP_BUILD_YOUR_OWN_SECTOR_MST_ID):
            # sector_rec = RP_create_sector_rec(sector_ID, sector_ID in sector_on, num_years, meta)
            sector_rec = RP_create_sector_rec(sector_ID, num_years, meta)
            # sector_rec = RP_create_sector_rec(sector_ID, sector_on[rpc.RP_SECTOR_CURR_ID_MAP[sector_ID] - 1] , num_years)
            sectors.append(sector_rec)

    return sectors

def RP_init_goal_records(sector_ID = int, num_years = int, meta : bool = False):
    goals = []

    for goal_ID in rpc.RP_SECTOR_GOALS[sector_ID]:
        goal_rec = RP_create_goal_rec(sector_ID, goal_ID, num_years, meta)
        goals.append(goal_rec)

    return goals

def RP_init_subgoal_records(sector_ID = int, goal_ID = int, num_years = int, meta : bool = False):
    subgoals = []

    for subgoal_ID in rpc.RP_GOAL_SUBGOALS[f"{sector_ID}_{goal_ID}"]:
        # user_spec_subgoal_ID = rpc.RP_GOAL_USER_SPEC_SUBGOALS[f"{sector_ID}_{goal_ID}"]

        # if meta or ((not meta) and (subgoal_ID == user_spec_subgoal_ID)):
        subgoal_rec = RP_create_subgoal_rec(sector_ID, goal_ID, subgoal_ID, num_years)
        subgoals.append(subgoal_rec)

    return subgoals

def RP_init_sector_list():
    sectors_on = deepcopy(rpc.RP_SECTOR_MST_IDS)
    sectors_on.remove(rpc.RP_BUILD_YOUR_OWN_SECTOR_MST_ID)
    return sectors_on
    # return np.full((rpc.RP_NUM_DEFAULT_SECTORS), True).tolist()

def RP_init_res_FP_scen_rec(include_status_quo = bool, value = 0.0):
    scenario_rec = {
        rpmvf.RP_RES_FP_SCEN_GOAL           : value,
        rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL : value
    }

    if include_status_quo:
        scenario_rec[rpmvf.RP_RES_FP_SCEN_STATUS_QUO] = value

    return scenario_rec

def RP_init_res_FP_scen_rec_by_year(use_lists = bool, scenarios_to_include : list = rpc.RP_FP_SCEN_LIST, num_years = int, value = 0.0):
    scenario_rec = {}

    if rpc.RP_FP_SCEN_STATUS_QUO_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_RES_FP_SCEN_STATUS_QUO] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if rpc.RP_FP_SCEN_GOAL_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_RES_FP_SCEN_GOAL] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if rpc.RP_FP_SCEN_COUNTERFACTUAL_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if use_lists:
        for key, value in scenario_rec.items():
            if isinstance(value, np.ndarray):
                scenario_rec[key] = value.tolist()

    return scenario_rec

def RP_init_res_FP_scen_rec_by_sex_year(use_lists = bool, include_status_quo = bool, num_years = int, value = 0.0):
    scenario_rec = {
        rpmvf.RP_RES_FP_SCEN_GOAL           : np.full((gbc.GB_NumSexes + 1, num_years), value, gbc.GB_CALC_FLOAT_TYPE),
        rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL : np.full((gbc.GB_NumSexes + 1, num_years), value, gbc.GB_CALC_FLOAT_TYPE)
    }

    if include_status_quo:
        scenario_rec[rpmvf.RP_RES_FP_SCEN_STATUS_QUO] = np.full((gbc.GB_NumSexes + 1, num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if use_lists:
        scenario_rec[rpmvf.RP_RES_FP_SCEN_GOAL] = scenario_rec[rpmvf.RP_RES_FP_SCEN_GOAL].tolist()
        scenario_rec[rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL] = scenario_rec[rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL].tolist()

        if include_status_quo:
            scenario_rec[rpmvf.RP_RES_FP_SCEN_STATUS_QUO] = scenario_rec[rpmvf.RP_RES_FP_SCEN_STATUS_QUO].tolist()

    return scenario_rec

def RP_init_res_FP_scen_rec_by_sex_age_year(use_lists = bool, num_years = int, value = 0.0):
    scenario_rec = {
        rpmvf.RP_RES_FP_SCEN_GOAL           : np.full((gbc.GB_NumSexes + 1, gbc.GB_MaxSingleAges + 1, num_years), value, gbc.GB_CALC_FLOAT_TYPE),
        rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL : np.full((gbc.GB_NumSexes + 1, gbc.GB_MaxSingleAges + 1, num_years), value, gbc.GB_CALC_FLOAT_TYPE)
    }

    if use_lists:
        scenario_rec[rpmvf.RP_RES_FP_SCEN_GOAL] = scenario_rec[rpmvf.RP_RES_FP_SCEN_GOAL].tolist()
        scenario_rec[rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL] = scenario_rec[rpmvf.RP_RES_FP_SCEN_COUNTERFACTUAL].tolist()

    return scenario_rec

def RP_init_FP_scen_scale_up_rec(num_years):
    return {
        rpmvf.RP_FP_SCEN_SC_ACCELERATED : np.full((num_years), 0.0, gbc.GB_CALC_FLOAT_TYPE).tolist(),
        rpmvf.RP_FP_SCEN_SC_MEDIAN      : np.full((num_years), 0.0, gbc.GB_CALC_FLOAT_TYPE).tolist(),
        rpmvf.RP_FP_SCEN_SC_FLATLINE    : np.full((num_years), 0.0, gbc.GB_CALC_FLOAT_TYPE).tolist(),
    }

def RP_init_res_SSP_cumu_rec():
    return {
        rpmvf.RP_SSP1 : 0,
        rpmvf.RP_SSP2 : 0,
        rpmvf.RP_SSP5 : 0,
    }

def RP_init_res_SSP_scen_rec_by_year(use_lists = bool, scenarios_to_include : list = rpc.RP_SSP_LIST, num_years = int, value = 0.0):
    scenario_rec = {}

    if rpc.RP_SSP_BASE_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_SSP_BASE] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if rpc.RP_SSP1_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_SSP1] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if rpc.RP_SSP2_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_SSP2] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if rpc.RP_SSP3_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_SSP3] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if rpc.RP_SSP5_MST_ID in scenarios_to_include:
        scenario_rec[rpmvf.RP_SSP5] = np.full((num_years), value, gbc.GB_CALC_FLOAT_TYPE)

    if use_lists:
        for key, value in scenario_rec.items():
            if isinstance(value, np.ndarray):
                scenario_rec[key] = value.tolist()

    return scenario_rec

def RP_init_sources_rec():
    return {
        rpmvf.RP_SRC_HH_SIZE                               : '',
        rpmvf.RP_SRC_FP_METHOD_MIX                         : '',
        rpmvf.RP_SRC_FP_EFFECTIVENESS                      : '',
        rpmvf.RP_SRC_FP_UNIT_COSTS                         : '',
        rpmvf.RP_SRC_DP_FIRST_YEAR_POP                     : '',
        rpmvf.RP_SRC_FP_TFR                                : '',

        rpmvf.RP_SRC_SECTOR_EDU_SCHOOL_AGE_RANGES          : '',
        rpmvf.RP_SRC_SECTOR_EDU_AVG_ANN_TEACHER_SAL_PRIM   : '',
        rpmvf.RP_SRC_SECTOR_EDU_ANN_COST_EDU_PRIM_STUDENT  : '',
        rpmvf.RP_SRC_SECTOR_EDU_ANN_COST_EDU_SEC_STUDENT    : '',

        rpmvf.RP_SRC_SECTOR_HLTH_AVG_UNIT_COST_SKILLED_DELIV     : '',
        rpmvf.RP_SRC_SECTOR_HLTH_AVG_UNIT_COST_ANC_4_PLUS_VISITS : '',
        rpmvf.RP_SRC_SECTOR_HLTH_PERC_SKILLED_HLTH_PROF_DOCTORS  : '',
        rpmvf.RP_SRC_SECTOR_HLTH_PERC_SKILLED_HLTH_PROF_NURSES   : '',
        rpmvf.RP_SRC_SECTOR_HLTH_AVG_SALARY_DOCTORS              : '',
        rpmvf.RP_SRC_SECTOR_HLTH_AVG_SALARY_NURSES               : '',

        rpmvf.RP_SRC_SECTOR_WASH_CAP_COSTS_PER_HH_IMPR_WATER     : '',
        rpmvf.RP_SRC_SECTOR_WASH_CAP_COSTS_PER_HH_IMPR_SANI      : '',

        rpmvf.RP_SRC_SECTOR_FOOD_PRIM_STAPLE_CROP                    : '',
        rpmvf.RP_SRC_SECTOR_FOOD_CROP_LAND                           : '',
        rpmvf.RP_SRC_SECTOR_FOOD_STAPLE_CROP_PROD_BASE               : '',
        rpmvf.RP_SRC_SECTOR_FOOD_TONNES_HARV_PERC_HECT               : '',
        rpmvf.RP_SRC_SECTOR_FOOD_ANN_KG_PER_YEAR_CONSUMP_STAPLE_CROP : '',
        rpmvf.RP_SRC_SECTOR_FOOD_PROD_COSTS_STAPLE_CROP              : '',
        rpmvf.RP_SRC_SECTOR_FOOD_DAILY_COST_FOOD_AID                 : '',

        rpmvf.RP_SRC_SECTOR_URB_PERC_POP_URB_BASE_END         : '',
        rpmvf.RP_SRC_SECTOR_URB_COST_PER_HH_MOVE_TO_HOUSING   : '',

        rpmvf.RP_SRC_SECTOR_ENERGY_CAP_COSTS_PER_HH_CONNECT_GRID    : '',
        rpmvf.RP_SRC_SECTOR_ENERGY_CAP_COSTS_PER_HH_CLEAN_COOKING   : '',

        rpmvf.RP_SRC_SECTOR_ECO_GDP_PER_CAPITA    : '',
        rpmvf.RP_SRC_SECTOR_ECO_LABOR_PARTIC_RATE : '',
        rpmvf.RP_SRC_SECTOR_ECO_CHANGE_GDP_CDR    : '',

        rpmvf.RP_SRC_GOALS_PRIMARY_ENROLL_ALL          : '',
        rpmvf.RP_SRC_GOALS_PRIMARY_ENROLL_MALE         : '',
        rpmvf.RP_SRC_GOALS_PRIMARY_ENROLL_FEMALE       : '',
        rpmvf.RP_SRC_GOALS_SECONDARY_ENROLL_ALL        : '',
        rpmvf.RP_SRC_GOALS_SECONDARY_ENROLL_MALE       : '',
        rpmvf.RP_SRC_GOALS_SECONDARY_ENROLL_FEMALE     : '',
        rpmvf.RP_SRC_GOALS_BIRTHS_ATTEND_SKILLED_PERS  : '',
        rpmvf.RP_SRC_GOALS_PREG_W_4_PLUS_ANC_VISITS    : '',
        rpmvf.RP_SRC_GOALS_HLTH_PROF_1000_POP          : '',
        rpmvf.RP_SRC_GOALS_PERC_ACCESS_IMP_WATER       : '',
        rpmvf.RP_SRC_GOALS_PERC_ACCESS_IMP_SANI        : '',
        rpmvf.RP_SRC_GOALS_PERC_SEV_FOOD_INSEC         : '',
        rpmvf.RP_SRC_GOALS_PERC_URB_POP_INFORM_SET     : '',
        rpmvf.RP_SRC_GOALS_PERC_POP_ACCESS_ELEC        : '',
        rpmvf.RP_SRC_GOALS_PERC_ACCESS_CLEAN_COOK_TECH : '',
        rpmvf.RP_SRC_GOALS_UNEMPLOYMENT_RATE           : '',
        rpmvf.RP_SRC_GOALS_PERC_NEET_ALL               : '',
        rpmvf.RP_SRC_GOALS_PERC_NEET_MALES             : '',
        rpmvf.RP_SRC_GOALS_PERC_NEET_FEMALES           : '',
    }

def RP_init_res_summary_rec():
    summary_rec = {
        rpmvf.RP_RES_SUM_CUMU_COSTS : {
            rpmvf.RP_RES_SUM_DIRECT   : 0.00,
            rpmvf.RP_RES_SUM_INDIRECT : 0.00,
        },

        rpmvf.RP_RES_SUM_CUMU_SAVINGS : {
            rpmvf.RP_RES_SUM_EDU_SECT_PRIM                    : 0.00,
            rpmvf.RP_RES_SUM_EDU_SECT_SEC                     : 0.00,
            rpmvf.RP_RES_SUM_HLTH_SECT_SKILLED_DELIV          : 0.00,
            rpmvf.RP_RES_SUM_HLTH_SECT_ANC                    : 0.00,
            rpmvf.RP_RES_SUM_WASH_SECT_IMPR_WATER             : 0.00,
            rpmvf.RP_RES_SUM_WASH_SECT_IMPR_SANI              : 0.00,
            rpmvf.RP_RES_SUM_FOOD_SEC_SECT_RED_FOOD_INSEC     : 0.00,
            rpmvf.RP_RES_SUM_URB_SECT_REDUCE_INFORM_SETT      : 0.00,
            rpmvf.RP_RES_SUM_ENERGY_SECT_INCR_ACC_ELEC        : 0.00,
            rpmvf.RP_RES_SUM_ENERGY_SECT_INCR_ACC_CLEAN_COOK  : 0.00,
        },

        rpmvf.RP_RES_SUM_SAVINGS_PER_DOLLAR_INVEST : 0.00,
    }

    return summary_rec


import uuid
from os import environ

from AvenirCommon.Database import GB_get_db_json

from SpectrumCommon.Const.GB import *
from SpectrumCommon.Const.PJ import *


def create_library_id():
    return uuid.uuid4().hex


def init_modvars(countryID, first_year, final_year, extra):
    subnatCode = extra.get(GB_Extra_SubnatCode, 0)
    subnatName = extra.get(GB_Extra_subnatName, '')

    projection_title = extra.get(GB_Extra_projection_title, '')
    modules = extra.get(GB_Extra_modules, '')
    author = extra.get(GB_Extra_Author, '')
    organization = extra.get(GB_Extra_organization, '')
    notes =  extra.get(GB_Extra_notes, '')
    appID = extra.get(GB_Extra_appID, '')
    library_id = create_library_id()

    LiST_costing_on = extra.get(GB_LIST_COSTING_ON, False)
    TB_costing_on   = extra.get(GB_TB_COSTING_ON, False)
    TB_statistical  = extra.get(GB_TB_ImpactStatistical, True)
    TB_dynamical    = extra.get(GB_TB_ImpactDynamical, False)
    # IHT costing is on if it was passed in and true or if both LiST and TB costing on are not on. This makes it so we
    # can be lazy and not have to specify IHT costing on for an IHT projection.
    if ((GB_IHT_ON in extra) and (extra[GB_IHT_ON])) or (not (LiST_costing_on or TB_costing_on)):
        IHT_costing_on = True
    else:
        IHT_costing_on = False

    if 'TBSubnatRegion' in extra:
        DBName = "subnationals\\TB_subnationalList_V2.JSON"
        tb_subnat_list = GB_get_db_json(environ["AVENIR_SW_DEFAULT_DATA_CONNECTION"], "tuberculosis", DBName)
        tb_subnat = tb_subnat_list[countryID][extra['TBSubnatRegion']] if countryID in tb_subnat_list else {}
    else:
        tb_subnat = {}

    #modvar creation list
    return {
            # db,
            #General
            PJN_TitleTag : projection_title,
            PJN_AppIDTag : appID,

            PJN_ProjectionValidTag: False,
            PJN_ProjectionSavedTag: False,
            PJN_AIMImpactCalculatedTag: False,
            PJN_UserSourceTag : [],
            #Versions
            PJN_VersionTag : GB_PM_Version,
            # PJN_VersionNumberTag :  GB_PM_VersionNumber,
            # PJN_BetaVersionTag :  GB_PM_BetaNumber,

            #years
            # PJN_BaseYearTag : GB_BaseYear,
            PJN_FirstYearTag : first_year,
            PJN_FinalYearTag : final_year,
        #     PJN_FinalCostingYearTag : final_costing_year,

            PJN_ReadOnlyTag : extra.get(GB_Extra_readonly, False),

            #country info
            PJN_CountryNameTag : extra['countryDetails']['name'],
            PJN_ISO3NumericTag : extra['countryDetails']['ISO3_Numeric'],
            PJN_ISO3AlphaTag : extra['countryDetails']['ISO3_Alpha'],
            PJN_SubNationalCodeTag : subnatCode,
            PJN_SubNationalNameTag : subnatName,

            PJN_IHTPlanTypeTag : extra.get(GB_Extra_plantype, ''),
            PJN_ModulesTag : modules,
            PJN_TBSubnatTag : tb_subnat,

            PJN_AuthorTag : author,
            PJN_OrganizationTag  : organization,
            PJN_NotesTag  : notes,
            # Note we need to set this library ID here instead of on "save"
            # because if we only do it on save if the user saves an open projection
            # then clicks save again we want to use the same library ID. If we only
            # generated it on save it wouldn't get passed on the 2nd save (with
            # current front end impl)
            PJN_LibraryIDTag: library_id,
            # Library version ID is the version of this projection in the library
            # initially empty but will be set on first save. Used for
            # optimistic locking of resource
            PJN_LibraryVersionIDTag: '',
            # PJN_Modes_Tag : {
            #     'LiSTCostingOn': LiST_costing_on,
            #     'TBCostingOn': TB_costing_on,
            #     'IHTOn': IHT_costing_on,
            #     'RAPIDSectorsOn':extra.get(GB_RAPID_SECTORS_ON, RP_init_sector_list()),
            #     'TBStatisticalOn':TB_statistical,
            #     'TBDynamicalOn':TB_dynamical,
            #     'GSKOn' : extra.get('GSKMode', False)
            # },

            "<PJN_ADRResourceID_V1>"    : '',
            "<PJN_ADRDataSetTitle_V1>"      : '',
    }






# fmt: off
import numpy as np
from os import environ as env

GBDatabaseVersion = 'V17'
# GBDatabaseVersion = 'Beta'
GBCountryListDBName = 'CountryList'

# version_str = lambda version, beta, special: f"{version} Beta {beta} {special}"

# Controller API Version number
GB_Controller_Version = f'Controller Server, 2025-02-26 Build 136 ({env.get("AVENIR_SW_MODE","NoMode")})'

# Calculation API Version number
GB_Calc_Version = f'Calculation Server, 2026-03-24 Build 336 ({env.get("AVENIR_SW_MODE","NoMode")})'

# projection manager API Version number.
GB_PM_Version = f'Projection Manager Server, 2026-03-20 Build 574 ({env.get("AVENIR_SW_MODE","NoMode")}) '

# Tool versions
# Extract Version number
GB_Extract_Version = "Extract Worker, 2025-10-17 Build 7"

# LiST SG Version number
GB_LiST_SG_Version = "LiST SG Worker, 2025-12-2 Build 14"

# AIM UA Version number
GB_AIM_UA_Version = "AIM UA Worker, 2025-12-18 Build 9"

# SAP Fitter Version number
GB_SAPFitter_Version = "SAP Worker, 2025-10-17 Build 13"

# Aggregate Version number
GB_Aggregate_Version = "Aggregate Worker, 2025-10-17 Build 5"

# TBDynamicalModel Version number
GB_TBDynamicalModel_Version = "TBDynamicalModel Worker, 2026-02-16 Build 22"

# Aggregate Version number
GB_AMFitLocalAdj_Version = "AM fit local adjustment Worker, 2025-10-17 Build 6"

# Multiproj Version number
GB_Multiproj_Version = "MultiProj Worker, 2025-10-17 Build 3"

# Data Import Version number
GB_IHTDataImport_Version = "IHT Data Import Worker, 2025-10-17 Build 5"

GB_DateTime_Format = "%Y-%m-%d %H:%M:%S"

# Projection file types for dialogs. *)
GB_AllPrjFiles = 1
GB_PJNZFiles = 2
GB_PJNFiles = 3
GB_ZIPFiles = 4
GB_XLSFiles = 5
GB_XLSXFiles = 6
GB_CSVFiles = 7
GB_MOFiles = 8
GB_EQFiles = 9
GB_EQQAFiles = 10
GB_GAMFiles = 1

# Modvar constants *)
MV_RW_ReadData = 1  # used when reading in a saved file normally
MV_RW_WriteData = 2  # used when saving data to file normally
MV_RW_ImportData = 3  # used when importing proj2 data into proj1
MV_RW_ExtendFinalYr = 4  # used when extending proj's final year
MV_RW_ReadLockedPrj = 5  # used for IHT modules
MV_RW_LoadDefData = 6  # used to load default values from a file
MV_RW_Comparison = 7  # used to load a comparison file that may have a
MV_InputVariable = 8
MV_ResultVariable = 9
MV_Extractable = 10
MV_General = 11
MV_DP = 12
MV_AM = 13
MV_HV = 14
MV_FP = 15
MV_DataListExtract = 16
MV_ConfigVariable = 17


GB_NativeYear = 1985
GB_BaseYear = 1970
GB_FirstPossibleYear = 1970

GB_Nan = "NAN"
GB_FloatDiff = 0.00001
GB_NOT_FOUND = -1

GB_BothSexes = 0
GB_Male = 1
GB_Female = 2
GB_NumSexes = GB_Female

GB_SexName = {
    GB_BothSexes : 'Both Sexes',
    GB_Male : 'Males',
    GB_Female : 'Females'
}

GB_January = 0
GB_February = 1
GB_March = 2
GB_April = 3
GB_May = 4
GB_June = 5
GB_July = 6
GB_August = 7
GB_September = 8
GB_October = 9
GB_November = 10
GB_December = 11

#   Inflation rate constants. Only need to hold values for local and US. Constants changed from desktop.
GB_NoInflation = 0
GB_InflationLocal = 1  # inflation for local currency
GB_InflationUS = 2  # inflation for USD
GB_NumInflationTypes = GB_InflationUS

#   Inflation method constants
GB_UseInflation = 1
GB_UseGDPDeflator = 2


#   Disability weight age groups
GB_DW_AllAges = 0
GB_DW_Age0_4 = 1
GB_DW_Age5_9 = 2
GB_DW_Age10_14 = 3
GB_DW_Age15_19 = 4
GB_DW_Age20_24 = 5
GB_DW_Age25_29 = 6
GB_DW_Age30_39 = 7
GB_DW_Age40_49 = 8
GB_DW_Age50_59 = 9
GB_DW_Age60_69 = 10
GB_DW_Age70_79 = 11
GB_DW_Age80_100 = 12
GB_DW_MaxAgeGrp = 12

# File types
GB_JSON = "JSON"

# PJNZ Read/Write col constants
GB_RW_TagCol = 0
GB_RW_DescriptCol = 1
GB_RW_NotesCol = 2
GB_RW_DataStartCol = 3

# Module codes (two-letter abbreviations) *)
GB_GBCode = "GB"
GB_DPCode = "DP"
GB_IVCode = "IV"
GB_FPCode = "FP"
GB_AMCode = "AM"
GB_RPCode = "RP"
GB_HVCode = "HV"
GB_CSCode = "CS"
GB_IHCode = "OH"
GB_HWCode = "HR"
GB_BACode = "BA"
GB_ICCode = "IC"
GB_ISCode = "IS"
GB_FSCode = "FS"
GB_PCCode = "PC"
GB_LGCode = "LG"
GB_TBCode = "TB"
GB_BGCode = "BG"
GB_HSCode = "HS"
GB_RNCode = "RN"
GB_GVCode = "GV"
GB_HFCode = "HF"
GB_NCCode = "NC"
GB_TICode = "TI"
GB_MACode = "MA"
GB_GCCode = "GC"
GB_HACode = "HA"
GB_MTCode = "MT"
GB_SICode = "SI"
GB_GTCode = "GT"
GB_PJCode = "PJ"
GB_VWCode = "VW"
GB_UACode = "UA"
GB_LCCode = "LC"
GB_TECode = "TE"


# Module ID's, do not change except for GB_MaxStdModules *)
# Module ID's are written to disk in the Projection files *)
GB_GB = 0  # {Main Shell}
GB_DP = 1
GB_IV = 2  # former GB_FC = 2     Focus (NewGen)
GB_FP = 3
GB_AM = 4
GB_RP = 5
GB_TROld = 6  # Former Training Module
GB_BCOld = 7  # Benefit Cost
GB_MTOld = 8  # MTCT
GB_LGOld = 9  # Logistics
GB_GLOld = 10  # Old Goals
GB_CROld = 11  # Condom Requirements
GB_SMOld = 12  # Safe Motherhood
GB_HV = 13
GB_ALOld = 14  # Allocate
GB_CS = 15
GB_IH = 16
GB_HW = 17
GB_BA = 18
GB_IC = 19
GB_IS = 20
GB_FS = 21
GB_PC = 22
GB_LG = 23
GB_TB = 24
GB_BG = 25
GB_HS = 26
GB_RN = 27
GB_GV = 28
GB_LOOld = 29  # Logistics optimization
GB_HF = 30
GB_NC = 31
GB_TI = 32
GB_MA = 33
GB_GC = 34
GB_HA = 35
GB_MT = 36
GB_SI = 37
GB_GT = 38
GB_PJ = 39
GB_VW = 40
GB_LF = 41
GB_MaxStdModules = 41

# Extras, pseudo modules (be sure these start at GB_MaxStdModules + 1)
GB_PseudoModuleStart = 100
GB_UA = 100
#  Even though LiST Costing is not truly a module and there is no separate
#  folder for it, treating it as a pseudo module makes certain processes
#  easier, like loading default data and creating projections with it in
#  MultiProj. TIME Econ is similar.
GB_LC = 101
GB_TE = 102
GB_NumPseudoModules = GB_TE - GB_PseudoModuleStart + 1

GB_PseudoModules = [GB_UA] #check for usage - prevents module inits from being run on projection creation

GB_MODULE_NUM_TO_MODULE_CODE = {
    0: GB_GBCode,
    1: GB_DPCode,
    2: GB_IVCode,
    3: GB_FPCode,
    4: GB_AMCode,
    5: GB_RPCode,
    13: GB_HVCode,
    15: GB_CSCode,
    16: GB_IHCode,
    17: GB_HWCode,
    18: GB_BACode,
    19: GB_ICCode,
    20: GB_ISCode,
    21: GB_FSCode,
    22: GB_PCCode,
    23: GB_LGCode,
    24: GB_TBCode,
    25: GB_BGCode,
    26: GB_HSCode,
    27: GB_RNCode,
    28: GB_GVCode,
    30: GB_HFCode,
    31: GB_NCCode,
    32: GB_TICode,
    33: GB_MACode,
    34: GB_GCCode,
    35: GB_HACode,
    36: GB_MTCode,
    37: GB_SICode,
    38: GB_GTCode,
    39: GB_PJCode,
    40: GB_VWCode,
    100: GB_UACode,
    101: GB_LCCode,
    102: GB_TECode,
}


# Modvar constants *)
MV_RW_ReadData = 1  # used when reading in a saved file normally
MV_RW_WriteData = 2  # used when saving data to file normally
MV_RW_ImportData = 3  # used when importing proj2 data into proj1
MV_RW_ExtendFinalYr = 4  # used when extending proj's final year
MV_RW_ReadLockedPrj = 5  # used for IHT modules
MV_RW_LoadDefData = 6  # used to load default values from a file
MV_RW_Comparison = 7  # used to load a comparison file that may have a
MV_InputVariable = 8
MV_ResultVariable = 9
MV_Extractable = 10
MV_General = 11
MV_DP = 12
MV_AM = 13
MV_HV = 14
MV_FP = 15
MV_DataListExtract = 16
MV_ConfigVariable = 17


GB_MOD_STR = {
    GB_GB: "GB",
    GB_DP: "DP",
    GB_IV: "IV",
    GB_FP: "FP",
    GB_AM: "AM",
    GB_RP: "RP",
    GB_HV: "HV",
    GB_CS: "CS",
    GB_IH: "OH",
    GB_HW: "HR",
    GB_BA: "BA",
    GB_IC: "IC",
    GB_IS: "IS",
    GB_FS: "FS",
    GB_PC: "PC",
    GB_LG: "LG",
    GB_TB: "TB",
    GB_BG: "BG",
    GB_HS: "HS",
    GB_RN: "RN",
    GB_GV: "GV",
    GB_HF: "HF",
    GB_NC: "NC",
    GB_TI: "TI",
    GB_MA: "MA",
    GB_GC: "GC",
    GB_HA: "HA",
    GB_MT: "MT",
    GB_SI: "SI",
    GB_GT: "GT",
    GB_PJ: "PJ",
    GB_VW: "VW",
    GB_UA: "UA",
    GB_LF: "LF",
}


GB_MOD_ID = {
    "GB": GB_GB,
    "DP": GB_DP,
    "IV": GB_IV,
    "FP": GB_FP,
    "AM": GB_AM,
    "RP": GB_RP,
    "HV": GB_HV,
    "CS": GB_CS,
    "OH": GB_IH,
    "HR": GB_HW,
    "BA": GB_BA,
    "IC": GB_IC,
    "IS": GB_IS,
    "FS": GB_FS,
    "PC": GB_PC,
    "LG": GB_LG,
    "TB": GB_TB,
    "BG": GB_BG,
    "HS": GB_HS,
    "RN": GB_RN,
    "GV": GB_GV,
    "HF": GB_HF,
    "NC": GB_NC,
    "TI": GB_TI,
    "MA": GB_MA,
    "GC": GB_GC,
    "HA": GB_HA,
    "MT": GB_MT,
    "SI": GB_SI,
    "GT": GB_GT,
    "PJ": GB_PJ,
    "UA": GB_UA,
}

GB_EXISTING_MOD_ID = {
    "GB": GB_GB,
    "PJ": GB_PJ,
    "DP": GB_DP,
    "AM": GB_AM,
    "FP": GB_FP,
    "TB": GB_TB,
    "CS": GB_CS,
    "OH": GB_IH,
    "IC": GB_IC,
    "PC": GB_PC,
    "HR": GB_HW,
    "IS": GB_IS,
    "HS": GB_HS,
    "FS": GB_FS,
    "BG": GB_BG,
    "UA": GB_UA,
    "RP": GB_RP,
    "GV": GB_GV,
    "HF": GB_HF,
    "LG": GB_LG,
    "HV": GB_HV,
    "RN": GB_RN,
}

GB_PSEUDO_MOD_ID = {
    "UA": GB_UA,
}

GB_NativeYear = 1970
GB_FirstPossibleYear = 1970

# Environment variables
GB_SPECT_MOD_DATA_CONN_ENV = "AVENIR_SW_DEFAULT_DATA_CONNECTION"

# Partial paths for DefaultData repo
GB_DEFAULT_DATA_SOURCE_DATA_PATH = r'DefaultData\SourceData'
GB_DEFAULT_DATA_JSON_DATA_PATH   = r'DefaultData\JSONData'
GB_DEFAULT_DATA_SCRIPTS_PATH     = r'DefaultData\Scripts'

# Container names
GB_GB_CONTAINER      = 'globals'
GB_IH_CORE_CONTAINER = 'ohtcore'
GB_IC_CONTAINER      = 'interventioncosting'
GB_IV_CONTAINER      = 'interventions'
GB_IS_CONTAINER      = 'infrastructure'
GB_LG_CONTAINER      = 'supplychain'
GB_HW_CONTAINER      = 'humanresources'
GB_CS_CONTAINER      = 'list'
GB_PC_CONTAINER      = 'programmecosting'
GB_BG_CONTAINER      = 'budgetmapping'
GB_TI_CONTAINER      = 'tuberculosis'
GB_RP_CONTAINER      = 'rapid'

GB_CONTAINERS = {
    GB_GB : GB_GB_CONTAINER,
    GB_IH : GB_IH_CORE_CONTAINER,
    GB_IC : GB_IC_CONTAINER,
    GB_IV : GB_IV_CONTAINER,
    GB_IS : GB_IS_CONTAINER,
    GB_HW : GB_HW_CONTAINER,
    GB_CS : GB_CS_CONTAINER,
    GB_PC : GB_PC_CONTAINER,
    GB_BG : GB_BG_CONTAINER,
    GB_TI : GB_TI_CONTAINER,
    GB_RP : GB_RP_CONTAINER,
    GB_LG : GB_LG_CONTAINER
}

# "extra" fields (passed into "extra" object when creating projections)
GB_IHT_ON                 = 'IHTCostingOn'
GB_TB_COSTING_ON          = 'TBCostingOn'
GB_LIST_COSTING_ON        = 'LiSTCostingOn'
GB_GSKMode                = 'GSKMode'
GB_MODULES                = 'modules'
# GB_RAPID_SECTORS_ON       = 'RAPIDSectorsOn'
GB_TB_ImpactStatistical   = 'TBImpactStatistical'
GB_TB_ImpactDynamical     = 'TBImpactDynamical'
GB_Extra_SubnatCode       = 'subNatCode'
GB_Extra_subnatName       = 'subNatName'
GB_Extra_projection_title = 'fileName'
GB_Extra_modules          = 'modules'
GB_Extra_Author           = 'author'
GB_Extra_organization     = 'organization'
GB_Extra_notes            = 'notes'
GB_Extra_appID            = 'appID'
GB_Extra_plantype         = 'planType'
GB_Extra_readonly         = 'readOnly'
GB_Extra_TBSubnatRegion   = 'TBSubnatRegion'
GB_Extra_epidemicStartYr  = 'epidemicStartYr'
GB_Extra_UseLeapfrog      = 'useLeapfrog'

# Treat as master IDs.
GB_NO_COSTING_MODE   = 0
GB_IHT_COSTING_MODE  = 1
GB_TB_COSTING_MODE   = 2
GB_LIST_COSTING_MODE = 3
GB_NUM_COSTING_MODES = GB_LIST_COSTING_MODE

GB_numpy_float     = np.float64
GB_numpy_flot_type = np.dtype(np.float64)
GB_CALC_FLOAT_TYPE = GB_numpy_float

# the first established subnational region code for any country will always be
# 10. Any subnational region code less than 10 will mean it is a custom region.
# if the subnational region code is 0, then it is the entire country
GB_FirstSubNatRegionCode       = 10
GB_EntireCountry               = 0
GB_CustomSubnatRegion          = 1

GB_PERCENT = 1
GB_NUMBER  = 2

GB_DYNAMIC = 1
GB_STATIC  = 2

# Single age constants
GB_MaxSingleAges         = 80

# 5-year age groups
GB_AllAges = 0
GB_A0_4    = 1
GB_A5_9    = 2
GB_A10_14  = 3
GB_A15_19  = 4
GB_A20_24  = 5
GB_A25_29  = 6
GB_A30_34  = 7
GB_A35_39  = 8
GB_A40_44  = 9
GB_A45_49  = 10
GB_A50_54  = 11
GB_A55_59  = 12
GB_A60_64  = 13
GB_A65_69  = 14
GB_A70_74  = 15
GB_A75_79  = 16
GB_A80_Up  = 17
GB_MAX_AGE = 17

# Interpolation types
GB_LINEAR          = 1
GB_S_SHAPED        = 2
GB_EXPONENTIAL     = 3
GB_FRONT_LOADED    = 4
GB_AgeCatStr = {

    GB_AllAges : 'AllAges',
    GB_A0_4    : 'Age 0-4',
    GB_A5_9    : 'Age 5-9',
    GB_A10_14  : 'Age 10-14',
    GB_A15_19  : 'Age 15-19',
    GB_A20_24  : 'Age 20-24',
    GB_A25_29  : 'Age 25-29',
    GB_A30_34  : 'Age 30-34',
    GB_A35_39  : 'Age 35-39',
    GB_A40_44  : 'Age 40-44',
    GB_A45_49  : 'Age 45-49',
    GB_A50_54  : 'Age 50-54',
    GB_A55_59  : 'Age 55-59',
    GB_A60_64  : 'Age 60-64',
    GB_A65_69  : 'Age 65-69',
    GB_A70_74  : 'Age 70-74',
    GB_A75_79  : 'Age 75-79',
    GB_A80_Up  : 'Age 80-Up',
}

# appIDs
GB_DP_Client      = 'DemProj'
GB_AIM_Client     = 'AIM'
GB_LiST_Client    = 'LiST'
GB_FamPlan_Client = 'FamPlan'
GB_Rapid_Client   = 'RAPID'
GB_IHT_Client     = 'IHT'
GB_TB_Client      = 'TB'
GB_Status_Client  = 'SWStatus'

GB_DP_Client_lc      = GB_DP_Client.lower()
GB_AIM_Client_lc     = GB_AIM_Client.lower()
GB_LiST_Client_lc    = GB_LiST_Client .lower()
GB_FamPlan_Client_lc = GB_FamPlan_Client.lower()
GB_Rapid_Client_lc   = GB_Rapid_Client.lower()
GB_IHT_Client_lc     = GB_IHT_Client .lower()
GB_TB_Client_lc      = GB_TB_Client.lower()
GB_Status_Client_lc  = GB_Status_Client.lower()

GB_AppList = [GB_DP_Client,
              GB_AIM_Client,
              GB_LiST_Client,
              GB_FamPlan_Client,
              GB_Rapid_Client,
              GB_IHT_Client,
              GB_TB_Client,
              GB_Status_Client]

gb_shape_module = {
    GB_DPCode : 'demproj',
    GB_AMCode : 'aim',
    GB_FPCode : 'famplan',
    GB_TBCode : 'tuberculosis',
}

GB_RestrictSaveProjList=[GB_LiST_Client, GB_TB_Client, GB_AIM_Client, GB_FamPlan_Client] # currently used to restrict saving

GB_AppLowerCaseList = [app.lower() for app in GB_AppList]

GB_SW_BETANOTALLOWED  = 'SW_BETANOTALLOWED'
GB_SW_NOTLOGGEDIN     = 'SW_NOTLOGGEDIN'
GB_SW_UNKNOWN         = 'SW_UNKNOWN'
GB_SW_GENERAL         = 'SW_GENERAL'
GB_SW_DOESNOTEXIST    = 'SW_DOESNOTEXIST'
GB_SW_INVALIDREQUEST  = 'SW_INVALIDREQUEST'
GB_SW_OLDERROR        = 'SW_OLDERROR'
GB_SW_REQUSUPERDET    = 'SW_REQUSUPERDET'
GB_SW_BADMODVARJSON   = 'SW_BADMODVARJSON'
GB_SW_FILTERFAILED    = 'SW_FILTERFAILED'
GB_SW_BADPROJPARAM    = 'SW_BADPROJPARAM'
GB_SW_BADGETMODVARS   = 'SW_BADGETMODVARS'
GB_SW_BADSETMODVARS   = 'SW_BADSETMODVARS'
GB_SW_BADCALCULATE    = 'SW_FAILEDCALCULATE'
GB_SW_BADEPPCALL      = 'SW_BADEPPCALL'
GB_SW_JWTEXPIRED      = 'SW_JWTEXPIRED'
GB_SW_JWTDECODEFAILED = 'SW_JWTDECODEFAILED'
GB_SW_THIRDPARTYFAILED = 'SW_THIRDPARTYFAILED'
GB_SW_NOPERMISSION     = 'SW_NOPERMISSION'
GB_SW_BUSY             = 'SW_BUSY'
GB_SW_MISSINGPROJECTION = 'SW_MISSINGPROJECTION'
GB_SW_PROJECTIONFILEMISSING = 'SW_PROJECTIONFILEMISSING'
GB_SW_PJNZIMPORTFAILED = "SW_PJNZIMPORTFAILED"
GB_SW_JSONIMPORTFAILED = "SW_JSONIMPORTFAILED"
GB_SW_PROJECTIONOUTOFDATE = 'SW_PROJECTIONOUTOFDATE'
GB_SW_PROJECTIONDELETED = 'SW_PROJECTIONOUTOFDATE'
GB_SW_PROJECTIONTOOMANY = 'SW_PROJECTIONTOOMANY'

GB_SW_ExceptMessages = {
    GB_SW_BETANOTALLOWED  : 'Operation not allowed in beta.',
    GB_SW_NOTLOGGEDIN     : 'Not logged in.',
    GB_SW_UNKNOWN         : 'Unknown error.',
    GB_SW_GENERAL         : 'General error.',
    GB_SW_INVALIDREQUEST  : 'Invalid request.',
    GB_SW_OLDERROR        : 'Old error exception.',
    GB_SW_BADMODVARJSON   : 'Failed to convert modvar to json.',
    GB_SW_DOESNOTEXIST    : 'The API call or link does not exist',
    GB_SW_REQUSUPERDET    : 'Requires admin priveleges',
    GB_SW_FILTERFAILED    : 'Filter failed.',
    GB_SW_BADPROJPARAM    : 'Bad projection parameters',
    GB_SW_BADCALCULATE    : 'Failed to calculate',
    GB_SW_BADGETMODVARS   : 'Failed to get modvars',
    GB_SW_BADSETMODVARS   : 'Failed to set modvars',
    GB_SW_BADEPPCALL      : 'Failed to make EPP call',
    GB_SW_JWTEXPIRED      : 'Login has expired, please login again.',
    GB_SW_JWTDECODEFAILED : 'Login failed, please try login again.',
    GB_SW_THIRDPARTYFAILED: 'Authorization failed. Please contact support.',
    GB_SW_NOPERMISSION    : 'You do not have permission to perform this action.',
    GB_SW_BUSY            : 'The server is currently busy. Please try again later.',
    GB_SW_MISSINGPROJECTION: 'Failed to locate projection.',
    GB_SW_PROJECTIONFILEMISSING: 'File missing from projection.',
    GB_SW_PJNZIMPORTFAILED: "Failed to import PJNZ.",
    GB_SW_JSONIMPORTFAILED: "Failed to import JSON.",
    GB_SW_PROJECTIONOUTOFDATE: "Failed to update projection. Someone else has updated this projection since you last opened it.",
    GB_SW_PROJECTIONDELETED: "Failed to update projection. Someone else has deleted since you last opened it.",
    GB_SW_PROJECTIONTOOMANY: "Failed to save projection, you have too many projections."
}
GB_SW_StatusCode = {
    GB_SW_BETANOTALLOWED  : 401,
    GB_SW_NOTLOGGEDIN     : 401,
    GB_SW_INVALIDREQUEST  : 422,
    GB_SW_UNKNOWN         : 500,
    GB_SW_GENERAL         : 500,
    GB_SW_OLDERROR        : 500,
    GB_SW_BADMODVARJSON   : 500,
    GB_SW_DOESNOTEXIST    : 404,
    GB_SW_REQUSUPERDET    : 401,
    GB_SW_FILTERFAILED    : 500,
    GB_SW_BADPROJPARAM    : 422,
    GB_SW_BADCALCULATE    : 500,
    GB_SW_BADGETMODVARS   : 500,
    GB_SW_BADSETMODVARS   : 500,
    GB_SW_BADEPPCALL      : 500,
    GB_SW_JWTEXPIRED      : 401,
    GB_SW_JWTDECODEFAILED : 401,
    GB_SW_THIRDPARTYFAILED: 401,
    GB_SW_NOPERMISSION    : 403,
    GB_SW_BUSY            : 429,
    GB_SW_MISSINGPROJECTION: 400,
    GB_SW_PROJECTIONFILEMISSING: 500,
    GB_SW_PJNZIMPORTFAILED: 400,
    GB_SW_JSONIMPORTFAILED: 400,
    GB_SW_PROJECTIONOUTOFDATE: 412,
    GB_SW_PROJECTIONDELETED: 412,
    GB_SW_PROJECTIONTOOMANY: 400,
}

# Languages
GB_ENGLISH_MST_ID    = 1
GB_FRENCH_MST_ID     = 2
GB_ARABIC_MST_ID     = 3
GB_SPANISH_MST_ID    = 4
GB_RUSSIAN_MST_ID    = 5
GB_PORTUGUESE_MST_ID = 6
GB_CHINESE_MST_ID    = 7
GB_INDONESIAN_MST_ID = 8
GB_UKRANIAN_MST_ID   = 9

GB_LANGUAGES = [
    GB_ENGLISH_MST_ID, GB_FRENCH_MST_ID, GB_ARABIC_MST_ID,
    GB_SPANISH_MST_ID, GB_RUSSIAN_MST_ID, GB_PORTUGUESE_MST_ID,
    GB_CHINESE_MST_ID, GB_INDONESIAN_MST_ID, GB_UKRANIAN_MST_ID
]

GB_STRING_DB_CURR_VERSION = 'V1'

GB_ENGLISH_STRINGS_DB_NAME = 'en'
GB_FRENCH_STRINGS_DB_NAME = 'fr'
GB_ARABIC_STRINGS_DB_NAME = ''
GB_SPANISH_STRINGS_DB_NAME = 'es'
GB_RUSSIAN_STRINGS_DB_NAME = 'ru'
GB_PORTUGUESE_STRINGS_DB_NAME = ''
GB_CHINESE_STRINGS_DB_NAME = ''
GB_INDONESIAN_STRINGS_DB_NAME = ''
GB_UKRANIAN_STRINGS_DB_NAME = 'uk'

GB_USER_GUEST_NAME = 'spectrum-guest'
GB_USER_GUEST_GUID = '5AA02053-2747-422F-A7B5-41A30194026B'
# USER_GUEST_GUID = '9C720571-1526-4EE3-9168-A2E6777103AD'

#Permission constants
GB_AllowBetaSave = "allowBetaSave"
GB_AllowGSK = "allowGSK"
GB_AllowSWQA = "allowSWQA"
# GB_AllowEQUIST = "allowEQUIST"



# Inputs

RP_VERSION_TAG                          = '<RP_Version_V1>'
RP_DATA_CHANGED_TAG                     = '<RP_DataChanged_V1>'
RP_PROJ_VALID_TAG                       = '<RP_ProjectionValid_V1>'
RP_FILE_NAME_TAG                        = '<RP_FileName_V1>'                        
RP_PROJ_VALID_DATE_TAG                  = '<RP_ProjectionValidDate_V1>'   

RP_SECTOR_RECORDS_TAG                   = '<RP_SectorRecords_V1>'
RP_SECTORS_ON_TAG                       = '<RP_RAPIDSectorsOn_V1>'

RP_FP_SCEN_GOAL_TAG                     = '<RP_FPScenarioGoal_V1>'
RP_FP_SCEN_GOAL_SCALE_UP_TAG            = '<RP_FPScenarioGoalScaleUp_V1>'
RP_FP_SCEN_COUNTERFACTUAL_TAG           = '<RP_FPScenarioCounterfactual_V1>'
RP_FP_SCEN_COUNTERFACTUAL_SCALE_UP_TAG  = '<RP_FPScenarioCounterfactualScaleUp_V1>'

RP_HOUSEHOLD_SIZE_BASELINE_TAG           = '<RP_HouseholdSizeBaseline_V1>' 

RP_FP_PERC_USERS_TAG                    = '<RP_FPPercUsers_V1>'
# RP_FP_TYPICAL_USE_EFFECTIVENESS_TAG     = '<RP_FPTypicalUseEffectiveness_V1>'
RP_FP_AVG_ANNUAL_UNIT_DIRECT_COSTS_TAG  = '<RP_FPAvgAnnualUnitDirectCosts_V1>'
RP_FP_INDIRECT_PROG_COSTS_TAG           = '<RP_FPIndirectProgramCosts_V1>'

RP_CLIMATE_PARAMS_TAG                  = '<RP_ClimateParameters_V1>'

RP_REVISED_SOURCES_TAG = '<RP_RevisedSources_V1>'

# Results

RP_RES_DP_BIG_POP_TAG                                 = '<RP_ResDPBigPop_V1>'
RP_RES_DP_TFR_TAG                                     = '<RP_ResDP_TFR_V1>'
RP_RES_FP_BIRTHS_TAG                                  = '<RP_ResFPBirths_V1>'

RP_RES_PRIM_SCH_STUDENTS_TAG                          = '<RP_ResPrimarySchoolStudents_V1>'
RP_RES_PRIM_EDU_COSTS_TAG                             = '<RP_ResPrimaryEducationCosts_V1>'
RP_RES_SEC_SCH_STUDENTS_TAG                           = '<RP_ResSecondarySchoolStudents_V1>'
RP_RES_SEC_EDU_COSTS_TAG                              = '<RP_ResSecondaryEducationCosts_V1>'
RP_RES_PRIM_SCH_TEACHERS_NEEDED_TAG                   = '<RP_ResPrimarySchoolTeachersNeeded_V1>'
RP_RES_PRIM_SCH_TEACHER_SALARY_COSTS_TAG              = '<RP_ResPrimaryTeacherSalaryCosts_V1>'

RP_RES_SKILLED_DELIV_TAG                              = '<RP_ResSkilledDeliveries_V1>'
RP_RES_SKILLED_DELIV_COSTS_TAG                        = '<RP_ResSkilledDeliveryCosts_V1>'
RP_RES_PREG_W_4_PLUS_ANC_VISITS_TAG                   = '<RP_ResPregnantWomen4PlusANCVisits_V1>'
RP_RES_COST_4_PLUS_ANC_VISITS_TAG                     = '<RP_ResCostOf4PlusANCVisits_V1>'
RP_RES_HLTH_PROF_TAG                                  = '<RP_ResHealthProfessionals_V1>'
RP_RES_HLTH_PROF_SALARY_COSTS_TAG                     = '<RP_ResHealthProfessionalSalaryCosts_V1>'

RP_RES_PEOPLE_ACCESS_IMRP_WATER_SOURCES_TAG           = '<RP_ResPeopleAccessImprWaterSources_V1>'
RP_RES_ADD_PEOPLE_NEED_ACCESS_IMPR_WATER_CUMU_TAG     = '<RP_ResAddPeopleNeedAccessImprWaterCumu_V1>'
RP_RES_CAP_COST_ADD_PEOPLE_GAIN_ACCESS_IMPR_WATER_TAG = '<RP_ResCapCostAddPeopleGainAccessImprWater_V1>'
RP_RES_PEOPLE_ACCESS_IMPR_SANI_FAC_TAG                = '<RP_ResPeopleAccessImprSaniFacilities_V1>'
RP_RES_ADD_PEOPLE_ACCESS_IMPR_SANI_FAC_CUMU_TAG       = '<RP_ResAddPeopleNeedAccessImprSaniFacilitiesCumu_V1>'
RP_RES_CAP_COST_ADD_PEOPLE_GAIN_ACCESS_IMPR_SANI_TAG  = '<RP_ResCapCostAddPeopleGainAccessImprSani_V1>'

RP_RES_MAIN_STAPLE_CROP_PROD_NEEDED_TAG               = '<RP_ResMainStapleCropProductionNeeded_V1>'
RP_RES_HECTARES_CROP_NEEDED_MEET_GOAL_TAG             = '<RP_ResHectaresCroplandNeededMeetGoal_V1>'
RP_RES_PERC_CROP_NEEDED_MEET_GOAL_TAG                 = '<RP_ResPercCroplandNeededMeetGoal_V1>'
RP_RES_COST_MAIN_STAPLE_CROP_PROD_TAG                 = '<RP_ResCostMainStapleCropProduction_V1>'
RP_RES_FOOD_NEEDED_RED_FOOD_INSEC_CUMU_TAG            = '<RP_ResFoodAidNeededReduceFoodInsecCumu_V1>'
RP_RES_COST_PROV_FOOD_AID_PEOPLE_FOOD_INSEC_ELIM_TAG  = '<RP_ResCostProvFoodAidPeopleFoodInsecElim_V1>'

RP_RES_URBAN_POP_TAG                                  = '<RP_ResUrbanPop_V1>'
RP_RES_URBAN_HHS_LIVING_INF_SETTLEMENTS_TAG           = '<RP_ResUrbanHHsLivingInformalSettlements_V1>'
RP_RES_HOUSES_TO_BUILD_CUMU_TAG                       = '<RP_ResHousesToBuildCumu_V1>'
RP_RES_COST_BUILD_HOUSES_PEOPLE_RELOC_TAG             = '<RP_ResCostBuildHousesPeopleRelocInfUrbanSettlements_V1>'

RP_RES_POP_ACCESS_ELEC_TAG                            = '<RP_ResPopAccessElectricity_V1>'
RP_RES_ADD_PEOPLE_NEED_ACCESS_ELEC_CUMU_TAG           = '<RP_ResAddPeopleNeedAccessElectricityCumu_V1>'
RP_RES_CAP_COST_CONNECT_ADD_HHS_ELEC_TAG              = '<RP_ResCapCostConnectAddHHsElectricity_V1>'
RP_RES_KWH_NEEDED_SUFF_ELEC_LEVELS_TAG                = '<RP_ResKWhNeededSufficientElectricityLevels_V1>'
RP_RES_COST_ELEC_TAG                                  = '<RP_ResCostElectricity_V1>'
RP_RES_HHS_CLEAN_COOK_FUEL_TECH_TAG                   = '<RP_ResHHsCleanCookingFuelTech_V1>'
RP_RES_ADD_HHS_NEED_ACCESS_CLEAN_COOK_TECH_CUMU_TAG   = '<RP_ResAddHHsNeedAccessCleanCookTechCumu_V1>'
RP_RES_COST_PROV_ADD_HHS_CLEAN_COOK_FUEL_TECH_TAG     = '<RP_ResCostProvAddHHsCleanCookFuelTech_V1>'

RP_RES_AGE_DEPENDENCY_RATIO_TAG                       = '<RP_ResAgeDependencyRatio_V1>'
RP_RES_GDP_PER_CAPITA_TAG                             = '<RP_ResGDPPerCapita_V1>'
RP_RES_TOTAL_PEOPLE_WITH_JOBS_TAG                     = '<RP_ResTotalPeopleWithJobs_V1>'
RP_RES_ADD_JOBS_NEEDED_TAG                            = '<RP_ResAddJobsNeeded_V1>'
RP_RES_NEET_TAG                                       = '<RP_ResNEET_V1>'
RP_RES_ADD_NEET_TAG                                   = '<RP_ResAddNEET_V1>'

RP_RES_NUM_POP_WITH_INTERV_TAG                        = '<RP_ResNumPopWithIntervention_V1>'
RP_RES_COST_POP_WITH_INTERV_TAG                       = '<RP_ResCostPopWithIntervention_V1>'

RP_RES_SSP_TEMP_CHANGES_FROM_BASE_TAG                 = '<RP_ResSSPTempChangesFromBaseline_V1>'
RP_RES_FEWER_KWH_NEEDED_BASED_ON_FP_ACC_TAG           = '<RP_ResFewerKWhNeededBasedFPAcceleration_V1>'
RP_RES_CUMU_SAVINGS_ELEC_COSTS_DUE_FP_ACC_TAG         = '<RP_ResCumulativeSavingsElectricityCostsDueFPAcc_V1>'
RP_RES_ADD_HECTARES_SAVED_CLIMATE_COND_TAG            = '<RP_ResAddHectaresSavedDueFPDiffClimateConditions_V1>'
RP_RES_GDP_PER_CAPITA_STATUS_QUO_TAG                  = '<RP_ResGDPPerCapitaStatusQuo_V1>'
RP_RES_GDP_PER_CAPITA_ACC_TAG                         = '<RP_ResGDPPerCapitaAccelerated_V1>'

RP_RES_SUMMARY_TAG                                    = '<RP_ResSummary_V1>'

# Meta ModVars. The values inside these do not change. They will automatically receive the latest default data when UpdateModvars is called. 

RP_SECTOR_RECORDS_META_TAG                   = '<RP_SectorRecords_Meta_V1>'

RP_FP_SCEN_GOAL_META_TAG                     = '<RP_FPScenarioGoal_Meta_V1>'
RP_FP_SCEN_COUNTERFACTUAL_META_TAG           = '<RP_FPScenarioCounterfactual_Meta_V1>'
RP_FP_SCEN_SCALE_UP_META_TAG                 = '<RP_FPScenarioScaleUp_Meta_V1>'

RP_HOUSEHOLD_SIZE_BASELINE_META_TAG          = '<RP_HouseholdSizeBaseline_Meta_V1>' 

RP_FP_PERC_USERS_META_TAG                    = '<RP_FPPercUsers_Meta_V1>'
RP_FP_TYPICAL_USE_EFFECTIVENESS_META_TAG     = '<RP_FPTypicalUseEffectiveness_Meta_V1>'
RP_FP_AVG_ANNUAL_UNIT_DIRECT_COSTS_META_TAG  = '<RP_FPAvgAnnualUnitDirectCosts_Meta_V1>'
RP_FP_INDIRECT_PROG_COSTS_META_TAG           = '<RP_FPIndirectProgramCosts_Meta_V1>'

RP_CLIMATE_PARAMS_META_TAG                   = '<RP_ClimateParameters_Meta_V1>'

RP_SSP_META_TAG                              = '<RP_SSP_Meta_V1>'

RP_SOURCES_META_TAG                          = '<RP_Sources_Meta_V1>'

# This is specifically singled out from all other sources because the client is running code that expects this particular format for every module.
# If that wasn't the case, this would just be a part of the sources meta MV above.
RP_RP_SOURCES_TAG                             = '<RP_RPSources_V1>'




# Intervention IDs contained in this file belong to interventions that have been removed/replaced in the 
# intervention database and do not appear in new projections. The IDs are referenced at least once in the 
# codebase and need to remain in the code so that we can support existing projections.

# The file containing current intervention IDs is overwritten when generating a new database, so these IDs 
# must be stored here in a separate file as opposed to alongside the current intervention IDs.

IC_IV_PulmTBCaseFindPatInitHIVNegChildren_MST_ID = '571'
IC_IV_PulmTBCaseFindPatInitHIVNegAdults_MST_ID = '570'
IC_IV_ExtraPulmTBCaseFindPatInitHIVNegChildren_MST_ID = '573'
IC_IV_ExtraPulmTBCaseFindPatInitHIVNegAdults_MST_ID = '572'


import SpectrumCommon.Const.GB.GBConst as gbc

import SpectrumCommon.Const.PC.PCModVarFields as pcmvf

#######################################################################################################
#                                                                                                     #
#                                       Cost categories                                               #
#                                                                                                     #
#######################################################################################################

def get_PC_cost_cat_rec(cost_cat_records = list, ID = str):
    cost_cat_recs = [cost_cat_rec for cost_cat_rec in cost_cat_records if cost_cat_rec[pcmvf.PC_COST_CAT_ID] == ID]
    return cost_cat_recs[0] if len(cost_cat_recs) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_cost_cat_idx(cost_cat_records = list, ID = str):
    indices = [idx for (idx, cost_cat_rec) in enumerate(cost_cat_records) if cost_cat_rec[pcmvf.PC_COST_CAT_ID] == ID]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND

### 

def get_PC_cost_cat_ID(cost_cat_rec = dict):
    return cost_cat_rec[pcmvf.PC_COST_CAT_ID]

def set_PC_cost_cat_ID(cost_cat_rec = dict, value = str):
    cost_cat_rec[pcmvf.PC_COST_CAT_ID] = value

###
   
def get_PC_cost_cat_name(cost_cat_rec = dict):
    return cost_cat_rec[pcmvf.PC_COST_CAT_NAME]

def set_PC_cost_cat_name(cost_cat_rec = dict, value = str):
    cost_cat_rec[pcmvf.PC_COST_CAT_NAME] = value

###
   
def get_PC_cost_cat_string_constant(cost_cat_rec = dict):
    return cost_cat_rec[pcmvf.PC_COST_CAT_STR_CONST]

def set_PC_cost_cat_string_constant(cost_cat_rec = dict, value = str):
    cost_cat_rec[pcmvf.PC_COST_CAT_STR_CONST] = value
 
###

def get_PC_cost_cat_custom(cost_cat_rec = dict):
    return cost_cat_rec[pcmvf.PC_COST_CAT_CUSTOM]

def set_PC_cost_cat_custom(cost_cat_rec = dict, value = bool):
    cost_cat_rec[pcmvf.PC_COST_CAT_CUSTOM] = value

###

#######################################################################################################
#                                                                                                     #
#                                       Cost inputs                                                   #
#                                                                                                     #
#######################################################################################################

def get_PC_num_cost_inputs(cost_input_records = list):
    return len(cost_input_records)

def get_PC_cost_input_rec(cost_input_records = list, ID = str):
    cost_input_recs = [cost_input_rec for cost_input_rec in cost_input_records if cost_input_rec[pcmvf.PC_COST_INPUT_ID] == ID]
    return cost_input_recs[0] if len(cost_input_recs) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_cost_input_idx(cost_input_records = list, ID = str):
    indices = [idx for (idx, cost_input_rec) in enumerate(cost_input_records) if cost_input_rec[pcmvf.PC_COST_INPUT_ID] == ID]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_cost_input_cat_ID(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_CAT_ID]

def set_PC_cost_input_cat_ID(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_CAT_ID] = value

###

def get_PC_cost_input_ID(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_ID]

def set_PC_cost_input_ID(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_ID] = value

###
   
def get_PC_cost_input_name(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_NAME]

def set_PC_cost_input_name(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_NAME] = value
 
###

def get_PC_cost_input_string_constant(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_STR_CONST]

def set_PC_cost_input_string_constant(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_STR_CONST] = value
 
###

def get_PC_cost_input_custom(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_CUSTOM]

def set_PC_cost_input_custom(cost_input_rec = dict, value = bool):
    cost_input_rec[pcmvf.PC_COST_INPUT_CUSTOM] = value

###


def get_PC_cost_input_unit_cost(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_UNIT_COST]

def set_PC_cost_input_unit_cost(cost_input_rec = dict, value = float):
    cost_input_rec[pcmvf.PC_COST_INPUT_UNIT_COST] = value

###

def get_PC_cost_input_unit_of_measure(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_UNIT_OF_MEASURE]

def set_PC_cost_input_unit_of_measure(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_UNIT_OF_MEASURE] = value

###

def get_PC_cost_input_unit_of_measure_string_constant(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_UNIT_OF_MEASURE_STR_CONST]

def set_PC_cost_input_unit_of_measure_string_constant(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_UNIT_OF_MEASURE_STR_CONST] = value

###

def get_PC_cost_input_note(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_NOTE]

def set_PC_cost_input_note(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_NOTE] = value

###

def get_PC_cost_input_source(cost_input_rec = dict):
    return cost_input_rec[pcmvf.PC_COST_INPUT_SOURCE]

def set_PC_cost_input_source(cost_input_rec = dict, value = str):
    cost_input_rec[pcmvf.PC_COST_INPUT_SOURCE] = value

###

# def get_PC_cost_input_source_string_constant(cost_input_rec = dict):
#     return cost_input_rec[pcmvf.PC_COST_INPUT_SOURCE_STR_CONST]

# def set_PC_cost_input_source_string_constant(cost_input_rec = dict, value = str):
#     cost_input_rec[pcmvf.PC_COST_INPUT_SOURCE_STR_CONST] = value

#######################################################################################################
#                                                                                                     #
#                                           Costing sections                                          #
#                                                                                                     #
#######################################################################################################

def get_PC_costing_sect_recs(costing_sect_records = list, area_type_ID = str, area_ID = str):
    costing_sect_recs = [costing_sect_rec for costing_sect_rec in costing_sect_records if \
        ((costing_sect_rec[pcmvf.PC_COSTING_SECT_AREA_TYPE_ID] == area_type_ID) \
        and (costing_sect_rec[pcmvf.PC_COSTING_SECT_AREA_ID] == area_ID))]
    return costing_sect_recs

def get_PC_num_costing_sects(costing_sect_records = list, area_type_ID = str, area_ID = str):
    costing_sect_records_reduced = get_PC_costing_sect_recs(costing_sect_records, area_type_ID, area_ID)
    if costing_sect_records_reduced == gbc.GB_NOT_FOUND:
        return 0
    else:
        return len(costing_sect_records_reduced)

def get_PC_costing_sect_rec(costing_sect_records: list, area_type_ID: str, area_ID: str, ID: str):
    costing_sect_recs = [costing_sect_rec for costing_sect_rec in costing_sect_records if \
        ((costing_sect_rec[pcmvf.PC_COSTING_SECT_AREA_TYPE_ID] == area_type_ID) \
        and (costing_sect_rec[pcmvf.PC_COSTING_SECT_AREA_ID] == area_ID) \
        and (costing_sect_rec[pcmvf.PC_COSTING_SECT_ID] == ID))]
    return costing_sect_recs[0] if len(costing_sect_recs) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_costing_sect_idx(costing_sect_records = list, area_type_ID = str, area_ID = str, ID = str):
    indices = [idx for (idx, costing_sect_rec) in enumerate(costing_sect_records) if \
        ((costing_sect_rec[pcmvf.PC_COSTING_SECT_AREA_TYPE_ID] == area_type_ID) \
        and (costing_sect_rec[pcmvf.PC_COSTING_SECT_AREA_ID] == area_ID) \
        and (costing_sect_rec[pcmvf.PC_COSTING_SECT_ID] == ID))]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_costing_sect_ID(costing_sect_rec = dict):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_ID]

def set_PC_costing_sect_ID(costing_sect_rec = dict, value = str):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_ID] = value

###
   
def get_PC_costing_sect_name(costing_sect_rec = dict):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_NAME]

def set_PC_costing_sect_name(costing_sect_rec = dict, value = str):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_NAME] = value

###

def get_PC_costing_sect_string_constant(costing_sect_rec = dict):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_STR_CONST]

def set_PC_costing_sect_string_constant(costing_sect_rec = dict, value = str):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_STR_CONST] = value

###

def get_PC_costing_sect_custom(costing_sect_rec = dict):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_CUSTOM]

def set_PC_costing_sect_custom(costing_sect_rec = dict, value = bool):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_CUSTOM] = value

###

def get_PC_costing_sect_costs_CS(costing_sect_rec = dict, t = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_COSTS_CS][t]

def set_PC_costing_sect_costs_CS(costing_sect_rec = dict, t = int, value = float):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_COSTS_CS][t] = value

###

def get_PC_costing_section_subsect_recs(costing_sect_rec = dict):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_SUBSECT_RECORDS]

###

def get_PC_costing_sect_price_elasticity_PPLI(costing_sect_rec = dict, col_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_PRICE_ELASTICITY_PPLI][col_curr_ID - 1]

def set_PC_costing_sect_price_elasticity_PPLI(costing_sect_rec = dict, col_curr_ID = int, value = float):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_PRICE_ELASTICITY_PPLI][col_curr_ID - 1] = value

###

def get_PC_costing_sect_income_elasticity_PPLI(costing_sect_rec = dict, col_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_INCOME_ELASTICITY_PPLI][col_curr_ID - 1]

def set_PC_costing_sect_income_elasticity_PPLI(costing_sect_rec = dict, col_curr_ID = int, value = float):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_INCOME_ELASTICITY_PPLI][col_curr_ID - 1] = value

###

def get_PC_costing_sect_price_income_only_PPLI(costing_sect_rec = dict, col_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_PRICE_INCOME_ONLY_PPLI][col_curr_ID - 1]

def set_PC_costing_sect_price_income_only_PPLI(costing_sect_rec = dict, col_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_PRICE_INCOME_ONLY_PPLI][col_curr_ID - 1] = value

###

def get_PC_costing_sect_non_price_measures_PPLI(costing_sect_rec = dict, col_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_NON_PRICE_MEASURES_PPLI][col_curr_ID - 1]

def set_PC_costing_sect_non_price_measures_PPLI(costing_sect_rec = dict, col_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_NON_PRICE_MEASURES_PPLI][col_curr_ID - 1] = value

###

def get_PC_costing_sect_all_MPOWER_PPLI(costing_sect_rec = dict, col_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_ALL_MPOWER_PPLI][col_curr_ID - 1]

def set_PC_costing_sect_all_MPOWER_PPLI(costing_sect_rec = dict, col_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_ALL_MPOWER_PPLI][col_curr_ID - 1] = value

###

def get_PC_costing_sect_tax_entry_mode_PPLI(costing_sect_rec = dict):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_ENTRY_MODE_PPLI]

def set_PC_costing_sect_tax_entry_mode_PPLI(costing_sect_rec = dict, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_ENTRY_MODE_PPLI] = value

###

def get_PC_costing_sect_real_retail_price_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_RETAIL_PRICE_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_real_retail_price_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_RETAIL_PRICE_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_excise_share_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_EXCISE_SHARE_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_excise_share_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_EXCISE_SHARE_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_tax_share_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_SHARE_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_tax_share_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_SHARE_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_real_exise_per_pack_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_EXCISE_PER_PACK_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_real_excise_per_pack_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_EXCISE_PER_PACK_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_real_tax_per_pack_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_TAX_PER_PACK_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_real_tax_per_pack_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_TAX_PER_PACK_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_real_industry_price_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_INDUSTRY_PRICE_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_real_industry_price_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_REAL_INDUSTRY_PRICE_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_tax_adjust_consump_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_ADJUST_CONSUMP_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_tax_adjust_consump_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_ADJUST_CONSUMP_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_excise_revenue_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_EXCISE_REVENUE_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_excise_revenue_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_EXCISE_REVENUE_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_tax_revenue_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_REVENUE_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_tax_revenue_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_TAX_REVENUE_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_usage_rate_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_USAGE_RATE_PPLI][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_usage_rate_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = float):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_USAGE_RATE_PPLI][year_endpoint_curr_ID - 1] = value

###

def get_PC_costing_sect_num_users_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int):
    return costing_sect_rec[pcmvf.PC_COSTING_SECT_NUM_USERS][year_endpoint_curr_ID - 1]

def set_PC_costing_sect_num_users_PPLI(costing_sect_rec = dict, year_endpoint_curr_ID = int, value = int):
    costing_sect_rec[pcmvf.PC_COSTING_SECT_NUM_USERS][year_endpoint_curr_ID - 1] = value

#######################################################################################################
#                                                                                                     #
#                                           Costing subsections                                       #
#                                                                                                     #
#######################################################################################################

def get_PC_costing_subsect_rec(costing_sect_records: list, area_type_ID: str, area_ID: str, costing_sect_ID: str, ID: str):
    costing_sect_rec = get_PC_costing_sect_rec(costing_sect_records, area_type_ID, area_ID, costing_sect_ID)
    if costing_sect_rec == gbc.GB_NOT_FOUND: return gbc.GB_NOT_FOUND
    costing_subsect_records = get_PC_costing_section_subsect_recs(costing_sect_rec)

    costing_subsect_recs = [costing_subsect_rec for costing_subsect_rec in costing_subsect_records if \
        (costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_ID] == ID)]
    return costing_subsect_recs[0] if len(costing_subsect_recs) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_costing_subsect_idx(costing_sect_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, ID = str):
    costing_sect_rec = get_PC_costing_sect_rec(costing_sect_records, area_type_ID, area_ID, costing_sect_ID)
    costing_subsect_records = get_PC_costing_section_subsect_recs(costing_sect_rec)

    indices = [idx for (idx, costing_subsect_rec) in enumerate(costing_subsect_records) if \
        (costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_ID] == ID)]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND

###
   
def get_PC_costing_subsect_name(costing_subsect_rec = dict):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_NAME]

def set_PC_costing_subsect_name(costing_subsect_rec = dict, value = str):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_NAME] = value
 
###

def get_PC_costing_subsect_string_constant(costing_subsect_rec = dict):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_STR_CONST]

def set_PC_costing_subsect_string_constant(costing_subsect_rec = dict, value = str):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_STR_CONST] = value
 
###

def get_PC_costing_subsect_ID(costing_subsect_rec = dict):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_ID]

def set_PC_costing_subsect_ID(costing_subsect_rec = dict, value = str):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_ID] = value

###

def get_PC_costing_subsect_custom(costing_subsect_rec = dict):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_CUSTOM]

def set_PC_costing_subsect_custom(costing_subsect_rec = dict, value = bool):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_CUSTOM] = value

###

def get_PC_costing_subsect_PPLI_phase_intensity(costing_subsect_rec = dict, row_type_curr_ID = int, t = int):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_PHASE_INTENSITY_VALUES_PPLI][row_type_curr_ID - 1][t]

def set_PC_costing_subsect_PPLI_phase_intensity(costing_subsect_rec = dict, row_type_curr_ID = int, t = int, value = float):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_PHASE_INTENSITY_VALUES_PPLI][row_type_curr_ID - 1][t] = value

###

def get_PC_costing_subsect_raise_taxes_price_PPLI(costing_subsect_rec = dict, t = int):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_RAISE_TAXES_PRICE_PPLI][t]

def set_PC_costing_subsect_raise_taxes_price_PPLI(costing_subsect_rec = dict, t = int, value = int):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_RAISE_TAXES_PRICE_PPLI][t] = value

###

def get_PC_costing_subsect_raise_taxes_excise_share_PPLI(costing_subsect_rec = dict, t = int):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_RAISE_TAXES_EXCISE_SHARE_PPLI][t]

def set_PC_costing_subsect_raise_taxes_excise_share_PPLI(costing_subsect_rec = dict, t = int, value = int):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_RAISE_TAXES_EXCISE_SHARE_PPLI][t] = value

###

def get_PC_costing_subsect_raise_taxes_tax_share_PPLI(costing_subsect_rec = dict, t = int):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_RAISE_TAXES_TAX_SHARE_PPLI][t]

def set_PC_costing_subsect_raise_taxes_tax_share_PPLI(costing_subsect_rec = dict, t = int, value = int):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_RAISE_TAXES_TAX_SHARE_PPLI][t] = value

###

def get_PC_costing_subsect_enter_total_inputs_per_admin_reg_year_PPLI(costing_subsect_rec = dict):
    return costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_ENTER_TOTAL_INPUTS_PER_ADMIN_REG_YEAR_PPLI]

def set_PC_costing_subsect_enter_total_inputs_per_admin_reg_year_PPLI(costing_subsect_rec = dict, value = bool):
    costing_subsect_rec[pcmvf.PC_COSTING_SUBSECT_ENTER_TOTAL_INPUTS_PER_ADMIN_REG_YEAR_PPLI] = value

#######################################################################################################
#                                                                                                     #
#                                           Activities                                                #
#                                                                                                     #
#######################################################################################################

def add_PC_activity_rec(activity_records = list, value = dict):
    activity_records.append(value)

def get_PC_act_rec(act_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, ID = str):
    act_recs = [act_rec for act_rec in act_records if \
        ((act_rec[pcmvf.PC_ACT_AREA_TYPE_ID] == area_type_ID) \
        and (act_rec[pcmvf.PC_ACT_AREA_ID] == area_ID) \
        and (act_rec[pcmvf.PC_ACT_COSTING_SECT_ID] == costing_sect_ID) \
        and (act_rec[pcmvf.PC_ACT_COSTING_SUBSECT_ID] == costing_subsect_ID) \
        and (act_rec[pcmvf.PC_ACT_ID] == ID))]
    return act_recs[0] if len(act_recs) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_act_idx(act_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, ID = str):
    indices = [idx for (idx, act_rec) in enumerate(act_records) if \
        ((act_rec[pcmvf.PC_ACT_AREA_TYPE_ID] == area_type_ID) \
        and (act_rec[pcmvf.PC_ACT_AREA_ID] == area_ID) \
        and (act_rec[pcmvf.PC_ACT_COSTING_SECT_ID] == costing_sect_ID) \
        and (act_rec[pcmvf.PC_ACT_COSTING_SUBSECT_ID] == costing_subsect_ID) \
        and (act_rec[pcmvf.PC_ACT_ID] == ID))]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_act_name(act_rec = dict):
    return act_rec[pcmvf.PC_ACT_NAME]

def set_PC_act_name(act_rec = dict, value = str):
    act_rec[pcmvf.PC_ACT_NAME] = value

###

def get_PC_act_string_constant(act_rec = dict):
    return act_rec[pcmvf.PC_ACT_STR_CONST]

def set_PC_act_string_constant(act_rec = dict, value = str):
    act_rec[pcmvf.PC_ACT_STR_CONST] = value

###

def get_PC_act_ID(act_rec = dict):
    return act_rec[pcmvf.PC_ACT_ID]

def set_PC_act_ID(act_rec = dict, value = str):
    act_rec[pcmvf.PC_ACT_ID] = value

###

def get_PC_act_custom(act_rec = dict):
    return act_rec[pcmvf.PC_ACT_CUSTOM]

def set_PC_act_custom(act_rec = dict, value = str):
    act_rec[pcmvf.PC_ACT_CUSTOM] = value

def get_PC_act_recs_for_costing_subsect(act_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str):
    act_recs = [act_rec for act_rec in act_records if ((act_rec[pcmvf.PC_ACT_AREA_TYPE_ID] == area_type_ID) \
        and (act_rec[pcmvf.PC_ACT_AREA_ID] == area_ID) and (act_rec[pcmvf.PC_ACT_COSTING_SECT_ID] == costing_sect_ID) \
        and (act_rec[pcmvf.PC_ACT_COSTING_SUBSECT_ID] == costing_subsect_ID))]
    return act_recs

###

def remove_PC_act_rec(act_records: list, area_type_ID: str, area_ID: str, costing_sect_ID: str, costing_subsect_ID: str, act_ID: str):
    act_recs = [act_rec for act_rec in act_records if not ((act_rec[pcmvf.PC_ACT_AREA_TYPE_ID] == area_type_ID) \
        and (act_rec[pcmvf.PC_ACT_AREA_ID] == area_ID) and (act_rec[pcmvf.PC_ACT_COSTING_SECT_ID] == costing_sect_ID) \
        and (act_rec[pcmvf.PC_ACT_COSTING_SUBSECT_ID] == costing_subsect_ID) and (act_rec[pcmvf.PC_ACT_ID] == act_ID))]
    return act_recs

###

# Remove all activities for the specified costing section
def remove_PC_acts_for_costing_sect(act_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str):
    act_recs = [act_rec for act_rec in act_records if not ((act_rec[pcmvf.PC_ACT_AREA_TYPE_ID] == area_type_ID) \
        and (act_rec[pcmvf.PC_ACT_AREA_ID] == area_ID) and (act_rec[pcmvf.PC_ACT_COSTING_SECT_ID] == costing_sect_ID))]
    return act_recs

###

# Remove all activities for the specified costing subsection
def remove_PC_acts_for_costing_subsect(act_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str):
    act_recs = [act_rec for act_rec in act_records if not ((act_rec[pcmvf.PC_ACT_AREA_TYPE_ID] == area_type_ID) \
        and (act_rec[pcmvf.PC_ACT_AREA_ID] == area_ID) and (act_rec[pcmvf.PC_ACT_COSTING_SECT_ID] == costing_sect_ID) \
        and (act_rec[pcmvf.PC_ACT_COSTING_SUBSECT_ID] == costing_subsect_ID))]
    return act_recs

#######################################################################################################
#                                                                                                     #
#                                           Activity line items                                       #
#                                                                                                     #
#######################################################################################################

def get_PC_num_ALIs(ALI_records = list):
    return len(ALI_records)

def get_PC_ALI_rec(ALI_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, act_ID = str, cost_input_ID = str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if \
        ((ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) \
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID) \
        and (ALI_rec[pcmvf.PC_ALI_ACT_ID] == act_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COST_INPUT_ID] == cost_input_ID))]
    return ALI_recs[0] if len(ALI_recs) > 0 else gbc.GB_NOT_FOUND
   
###
   
def get_PC_ALI_idx(ALI_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, act_ID = str, cost_input_ID = str):
    indices = [idx for (idx, ALI_rec) in enumerate(ALI_records) if \
        ((ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) \
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID) \
        and (ALI_rec[pcmvf.PC_ALI_ACT_ID] == act_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COST_INPUT_ID] == cost_input_ID))]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND

###

def get_PC_ALI_area_type_ID(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID]

def set_PC_ALI_area_type_ID(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] = value

###

def get_PC_ALI_area_ID(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_AREA_ID]

def set_PC_ALI_area_ID(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_AREA_ID] = value


###

def get_PC_ALI_costing_sect_ID(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID]

def set_PC_ALI_costing_sect_ID(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] = value

###

def get_PC_ALI_costing_subsect_ID(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID]

def set_PC_ALI_costing_subsect_ID(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] = value

###

def get_PC_ALI_act_ID(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_ACT_ID]

def set_PC_ALI_act_ID(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_ACT_ID] = value

###

def get_PC_ALI_cost_input_ID(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_COST_INPUT_ID]

def set_PC_ALI_cost_input_ID(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_COST_INPUT_ID] = value

###

def get_PC_ALI_num_units(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_NUM_UNITS]

def set_PC_ALI_num_units(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_NUM_UNITS] = value

###

def get_PC_ALI_num_units_PPLI(ALI_rec = dict, admin_level_curr_ID = int):
    return ALI_rec[pcmvf.PC_ALI_NUM_UNITS][admin_level_curr_ID - 1]

def set_PC_ALI_num_units_PPLI(ALI_rec = dict, admin_level_curr_ID = int, value = int):
    ALI_rec[pcmvf.PC_ALI_NUM_UNITS][admin_level_curr_ID - 1] = value

###

def get_PC_ALI_duration(ALI_rec = dict):
    return ALI_rec[pcmvf.PC_ALI_DURATION]

def set_PC_ALI_duration(ALI_rec = dict, value = int):
    ALI_rec[pcmvf.PC_ALI_DURATION] = value

###

def get_PC_ALI_duration_PPLI(ALI_rec = dict, admin_level_curr_ID = int):
    return ALI_rec[pcmvf.PC_ALI_DURATION][admin_level_curr_ID - 1]

def set_PC_ALI_duration_PPLI(ALI_rec = dict, admin_level_curr_ID = int, value = int):
    ALI_rec[pcmvf.PC_ALI_DURATION][admin_level_curr_ID - 1] = value

###

def get_PC_ALI_frequency(ALI_rec = dict, t = int):
    return ALI_rec[pcmvf.PC_ALI_FREQUENCY][t]

def set_PC_ALI_frequency(ALI_rec = dict, t = int, value = int):
    ALI_rec[pcmvf.PC_ALI_FREQUENCY][t] = value

###

def get_PC_ALI_quantity(ALI_rec = dict, ALI_set_curr_ID = int, admin_level_curr_ID = int, t = int):
    return ALI_rec[pcmvf.PC_ALI_QUANTITY_PER_ADMIN_REG_YEAR_PPLI][ALI_set_curr_ID - 1][admin_level_curr_ID - 1][t]

def set_PC_ALI_quantity(ALI_rec = dict, ALI_set_curr_ID = int, admin_level_curr_ID = int, t = int, value = int):
    ALI_rec[pcmvf.PC_ALI_QUANTITY_PER_ADMIN_REG_YEAR_PPLI][ALI_set_curr_ID - 1][admin_level_curr_ID - 1][t] = value

###

def add_PC_ALI_rec(ALI_records = list, value = dict):
    ALI_records.append(value)

###

# Return all activity line items for the specified costing section
def get_PC_ALIs_for_costing_sect(ALI_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if ((ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) and \
        (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID))]
    return ALI_recs

###

# Return all activity line items for the specified costing subsection
def get_PC_ALIs_for_costing_subsect(ALI_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if (
        (ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) 
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) 
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID) 
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID)
    )]
    return ALI_recs

###

# Return all activity line items for the specified activity
def get_PC_ALIs_for_act(ALI_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, act_ID = str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if ((ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) \
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID) and (ALI_rec[pcmvf.PC_ALI_ACT_ID] == act_ID))]
    return ALI_recs

###

# Return all activity line items for the specified activity and cost category.
def get_PC_ALIs_for_act_and_cost_cat(ALI_records = list, cost_input_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, act_ID = str, cost_cat_ID = str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if (
        (ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID)
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) 
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID) 
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID) 
        and (ALI_rec[pcmvf.PC_ALI_ACT_ID] == act_ID) 
        and (get_PC_cost_input_cat_ID(get_PC_cost_input_rec(cost_input_records, ALI_rec[pcmvf.PC_ALI_COST_INPUT_ID])) == cost_cat_ID)
    )]
    return ALI_recs

###

# Remove all activity line items for the specified costing section
def remove_PC_ALIs_for_costing_sect(ALI_records: list, area_type_ID: str, area_ID: str, costing_sect_ID: str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if not ((ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) and \
        (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID))]
    return ALI_recs

###

# Remove all activity line items for the specified costing subsection
def remove_PC_ALIs_for_costing_subsect(ALI_records: list, area_type_ID: str, area_ID: str, costing_sect_ID: str, costing_subsect_ID: str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if not ((ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) \
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID) \
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID))]
    return ALI_recs

# Remove all activity line items for the specified activity
def remove_PC_ALIs_for_act(ALI_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, act_ID = str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if (
        not ((ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID) 
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID) 
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID) 
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID) 
        and (ALI_rec[pcmvf.PC_ALI_ACT_ID] == act_ID))
    )]
    return ALI_recs

# Remove all activity line items associated with the cost input
def remove_PC_ALIs_for_cost_input(ALI_records = list, cost_input_ID = str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if not ((ALI_rec[pcmvf.PC_ALI_COST_INPUT_ID] == cost_input_ID))]
    return ALI_recs

def remove_PC_ALI_rec(ALI_records: list, area_type_ID: str, area_ID: str, costing_sect_ID: str, costing_subsect_ID: str, act_ID: str, cost_input_ID: str):
    ALI_recs = [ALI_rec for ALI_rec in ALI_records if not (
        (ALI_rec[pcmvf.PC_ALI_AREA_TYPE_ID] == area_type_ID)
        and (ALI_rec[pcmvf.PC_ALI_AREA_ID] == area_ID)
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SECT_ID] == costing_sect_ID)
        and (ALI_rec[pcmvf.PC_ALI_COSTING_SUBSECT_ID] == costing_subsect_ID)
        and (ALI_rec[pcmvf.PC_ALI_ACT_ID] == act_ID)
        and (ALI_rec[pcmvf.PC_ALI_COST_INPUT_ID] == cost_input_ID)
    )]
    return ALI_recs

#######################################################################################################
#                                                                                                     #
#                                           Results                                                   #
#                                                                                                     #
#######################################################################################################

def get_PC_costs_by_sect(costs_by_sect_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str):
    recs = [rec for rec in costs_by_sect_records if (
        (rec[pcmvf.PC_F_AREA_TYPE_ID] == area_type_ID) 
        and (rec[pcmvf.PC_F_AREA_ID] == area_ID) 
        and (rec[pcmvf.PC_F_COSTING_SECT_ID] == costing_sect_ID) 
    )]
    # There will be either one or zero entries.
    return recs[0][pcmvf.PC_F_VALUE] if len(recs) > 0 else gbc.GB_NOT_FOUND

###
 
def get_PC_costs_by_subsect(costs_by_subsect_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str):
    recs = [rec for rec in costs_by_subsect_records if (
        (rec[pcmvf.PC_F_AREA_TYPE_ID] == area_type_ID) 
        and (rec[pcmvf.PC_F_AREA_ID] == area_ID) 
        and (rec[pcmvf.PC_F_COSTING_SECT_ID] == costing_sect_ID) 
        and (rec[pcmvf.PC_F_COSTING_SUBSECT_ID] == costing_subsect_ID) 
    )]
    # There will be either one or zero entries.
    return recs[0][pcmvf.PC_F_VALUE] if len(recs) > 0 else gbc.GB_NOT_FOUND

###

def get_PC_costs_by_act(costs_by_act_records = list, area_type_ID = str, area_ID = str, costing_sect_ID = str, costing_subsect_ID = str, act_ID = str):
    recs = [rec for rec in costs_by_act_records if (
        (rec[pcmvf.PC_F_AREA_TYPE_ID] == area_type_ID) 
        and (rec[pcmvf.PC_F_AREA_ID] == area_ID) 
        and (rec[pcmvf.PC_F_COSTING_SECT_ID] == costing_sect_ID) 
        and (rec[pcmvf.PC_F_COSTING_SUBSECT_ID] == costing_subsect_ID) 
        and (rec[pcmvf.PC_F_ACT_ID] == act_ID)
    )]
    # There will be either one or zero entries.
    return recs[0][pcmvf.PC_F_VALUE] if len(recs) > 0 else gbc.GB_NOT_FOUND


from SpectrumCommon.Const.DP import *
from SpectrumCommon.Const.GB import *
import numpy as np

def getIRRFittingDataDict():
    return {
        'AIC'       : 0.0,
        'isFitted'  : False,
        'date'      : DPNotAvail,
    }

def getDefaultEpidemicConst(countryData):
    result = DP_Concentrated
    if 'EpidemicType' in countryData:
        epidemicType = countryData['EpidemicType']['Epidemic Type']
        if epidemicType == 'Generalized':
            result = DP_Generalized
        elif epidemicType == 'Concentrated':
            result = DP_LowLevel
    return result

def getPrevSurveyDict(calcYear):
    result = {
        'data' : np.zeros((GB_Female + 1, DP_WeightedSampleSize + 1, DP_A80_Up + 1)).tolist(),
        'used' : False,
        'name' : '',
        'year' : max(calcYear, DP_DefaultSurveyYear),
    }
    return result

def getARTCovSurveyDict(calcYear):
    result = {
        'data' : np.zeros((GB_Female + 1, DP_WeightedSampleSize + 1, DP_Val_A50Plus + 1)).tolist(),
        'used' : False,
        'name' : '',
        'year' : max(calcYear, DP_DefaultSurveyYear),
    }
    return result
    
def getCustomSAPDataDict(finalYearIdx):
    result = {
        'name' : '',  
        'id' : DPNotAvail,  
        'dataSource' : DPNotAvail,                   # IncEpidemicRGIdx
        'fitType' : DPNotAvail,                      # SAPFitType
        'HIVPrevModel' : DPNotAvail,                 # HIVPrevModel
        'AIC' : DPNotAvail,
        'HIVSexRatio' : np.zeros((finalYearIdx + 1)).tolist(),
        'DistOfHIV' : np.zeros((GB_Female + 1, DP_MAX_AGE + 1, finalYearIdx + 1)).tolist(),
    }
    return result







from SpectrumCommon.Const.DP import *
from SpectrumCommon.Modvars.DP.DPUtil import *
from SpectrumCommon.Const.GB import *
from SpectrumCommon.Modvars.GB.GBDefs import *
from AvenirCommon.Util import GBRange, getDataYear
from SpectrumCommon.Modvars.GB.GBUtil import  getYearIdx, getYear
from SpectrumCommon.Modvars.DP.DPUtil import sexMap

import numpy as np

# def EdAdSource_Init():
#     result = []
#     # (* Data Source Variables *)
#     # result : TstringList.Create
#     # 'Sorted : False
#     return result

def EdDmSource_Init():
    result = []
    # result : TstringList.Create
    # 'Sorted : False
    return result

def DPSources_Init(UPDdb):
    def getSourceDict():
        return {
            'Name' : 'N/A',
            'Summary' : 'N/A',
            'Source' : 'N/A',
            'Date' : 'N/A',
        }

    UPD = {
        'Provider' : '',
        'VersionNum' : 0,
        'Sources' : [getSourceDict() for i in GBRange(DP_UPD_Pop, DP_UPD_Migration)]
    }

    if 'general' in UPDdb:
        UPD['Provider'] = UPDdb['general']['Provider']
        UPD['VersionNum'] = UPDdb['general']['Version']
    
    if 'source' in UPDdb:   
        for src in GBRange(DP_UPD_Pop, DP_UPD_Migration):
            UPD['Sources'][src]['Name'] = UPDdb['source'][src]['Name']
            UPD['Sources'][src]['Summary'] = UPDdb['source'][src]['Summary']
            UPD['Sources'][src]['Source'] = UPDdb['source'][src]['Source']
            UPD['Sources'][src]['Date'] = UPDdb['source'][src]['RefDate']

    return UPD

def DefaultUPDLE_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear

    result = getList_sex_year(projInfo['finalYrIdx'])
    if 'lfts' in UPDdb:
        for sex in [GB_Male, GB_Female]:
            for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
                dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
                result[sex][t] = UPDdb['lfts'][dbyear][0][sex]['ex']

    return result

def DefaultUPDIMR_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear

    result = getList_sex_year(projInfo['finalYrIdx'])

    if 'lfts' in UPDdb:
        for sex in [GB_Male, GB_Female]:
            for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
                dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
                result[sex][t] = (1 - (UPDdb['lfts'][dbyear][1][sex]['lx'] / 100000)) * 1000

    return result

def DefaultUPDCMR_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear

    result = getList_sex_year(projInfo['finalYrIdx'])

    if 'lfts' in UPDdb:
        for sex in [GB_Male, GB_Female]:
            for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
                dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
                result[sex][t] = (1 - (UPDdb['lfts'][dbyear][5][sex]['lx'] / 100000)) * 1000
 
    return result

def CustomFilePYears_Init():
    result = {
        'DP_CPopFirstYr' : -1,
        'DP_CPopFinalYr' : -1,
    }
    return result

def CustomPopStopRescalingYear_Init(projInfo):
    result = projInfo['finalYr']
    return result

def DefaultUPDSurvRate_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear

    result = np.zeros((DP_MAX_SURVRATE + 1, GB_Female + 1, projInfo['finalYrIdx'] + 1)).tolist()

    if 'lfts' in UPDdb:
        for age in GBRange(0, DP_MaxSingleAges + 1):
            for sex in [GB_Male, GB_Female]:
                for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
                    dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
                    result[age + 1][sex][t] = UPDdb['lfts'][dbyear][age][sex]['sx']

    return result

def BigPop_Init(projInfo, UPDdb, InitialConditionsDB):
    result = getList_sex_singleAge_year(projInfo['finalYrIdx'])
    firstYear = projInfo['firstYr']
    firstYearIdx = getYearIdx(int(firstYear), projInfo['firstYr'])

    # if the first year is 1970, use the values directly from the UPD files.  
    # this is extra important for regeneration of the InitialConditionsDB so we don't compound any errors that could occur
    # in a closed loop system.  freshly seeding the base year data is good and stuff 
    if firstYear == GB_BaseYear:  
        if 'basepop' in UPDdb:                                          
            for age in GBRange(0, DP_MaxSingleAges):
                for sex in GBRange(GB_Male, GB_Female):
                            result[sex][age][firstYearIdx] = UPDdb['basepop'][age][sex]
    else: 
        Pop1DefaultDBArr = np.array(InitialConditionsDB['pop1'])
        for age in GBRange(0, DP_MaxSingleAges):
            for sex in GBRange(GB_Male, GB_Female):
                result[sex][age][firstYearIdx] = Pop1DefaultDBArr[firstYear - GB_BaseYear, sexMap[sex], age, :, :].sum() #t, S, A, CD4, Dur
    
    result[GB_BothSexes] = np.array(result)[GBRange(GB_Male, GB_Female)].sum(axis=0).tolist()

    return result

def FirstYearPop_Meta_Init(projInfo, UPDdb, InitialConditionsDB, subnat):
    return np.array(BigPop_Init(projInfo, UPDdb, InitialConditionsDB, subnat))[:, :, 0].tolist()
    
def DP_InitialConditions_Init(projInfo, InitialConditionsDB):
    firstYear = projInfo['firstYr']
    result = {
        'pop1'  : InitialConditionsDB['pop1'][firstYear - GB_BaseYear] if (not (firstYear == GB_BaseYear)) else 'why are you here',
        'births'  : InitialConditionsDB['births'][firstYear - GB_BaseYear] if (not (firstYear == GB_BaseYear)) else 'why are you here',
        'deaths'  : InitialConditionsDB['deaths'][firstYear - GB_BaseYear] if (not (firstYear == GB_BaseYear)) else 'why are you here',
    }

    return result

def TFR_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear
 
    result = getList_year(projInfo['finalYrIdx'])

    if 'tfr' in UPDdb:
        for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
            dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
            result[t] = UPDdb['tfr'][dbyear]

    return result

def ASFR_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear
    
    result = getList_5yrAge_year(projInfo['finalYrIdx'])

    if 'pasfrs' in UPDdb:
        for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
            dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
            sum = 0
            for age in GBRange(15, 49):
                result[singleAgeTo5YearAG(age)][t] += UPDdb['pasfrs'][dbyear][age] * 100
                sum += UPDdb['pasfrs'][dbyear][age] * 100
            result[DP_AllAges][t] = sum

    return result

def UNPopASFR_Init(projInfo, UPDdb):   
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear
    
    result = getList_5yrAge_year(projInfo['finalYrIdx'])

    if 'pasfrs' in UPDdb:
        for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
            dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
            sum = 0
            for age in GBRange(15, 49):
                result[singleAgeTo5YearAG(age)][t] += UPDdb['pasfrs'][dbyear][age] * 100
                sum += UPDdb['pasfrs'][dbyear][age] * 100
            result[DP_AllAges][t] = sum

    return result

def ASFRTables_Init(DPGlobalData):

    result = np.zeros((DP_ASFR_Average + 1, DP_ASFR_NumCols + 1, DP_ASFR_NumRows + 1)).tolist()

    for table in GBRange(DP_ASFR_UNAfrica, DP_ASFR_Average):
        tableName = ASFRTables[table]
        for row in GBRange(1, DP_ASFR_NumRows):
            for col in GBRange(1, DP_ASFR_NumCols): 
                result[table][col][row] = DPGlobalData['ASFRTables'][tableName][row - 1][col - 1]

    return result

def SexBirthRatio_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear

    result = getList_year(projInfo['finalYrIdx'])

    if 'srb' in UPDdb:
        for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
            dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
            result[t] = UPDdb['srb'][dbyear]

    return result

def LE_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear

    result = getList_sex_year(projInfo['finalYrIdx'])

    if 'lfts' in UPDdb:
        for sex in [GB_Male, GB_Female]:
            for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
                dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
                result[sex][t] = UPDdb['lfts'][dbyear][0][sex]['ex']

    return result

def MigrRate_Init(projInfo, UPDdb):
    dataFirstYear = 1970
    dataFinalYear = DP_UPDFinalYear
    
    result = getList_sex_5yrAge_year(projInfo['finalYrIdx'])

    if 'migration' in UPDdb:
        for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
            dbyear = str(getDataYear(dataFirstYear, dataFinalYear, getYear(t, projInfo['firstYr'])))   #for flatlining
            for sex in GBRange(GB_Male, GB_Female):
                sum = 0
                for age in GBRange(0, 80):
                    result[sex][singleAgeTo5YearAG(age)][t] += UPDdb['migration'][dbyear][age][sex]
                    sum += UPDdb['migration'][dbyear][age][sex]
                result[sex][DP_AllAges][t] = sum

    return result

def MigrAgeDist_Init(projInfo, migrRate):
    result = getList_sex_5yrAge_year(projInfo['finalYrIdx'])

    for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
        for sex in GBRange(GB_Male, GB_Female):
            for age in GBRange(DP_A0_4, DP_MAX_AGE):
                if migrRate[sex][DP_AllAges][t] == 0:
                    result[sex][age][t] = 0
                else:           
                    result[sex][age][t] = migrRate[sex][age][t] / migrRate[sex][DP_AllAges][t] * 100

    return result

def COVID19DeathAgeDist_Init(projInfo, DPGlobalData):

    result = getList_sex_5yrAge_year(projInfo['finalYrIdx'])

    for t in GBRange(DP_MIN_YEAR, projInfo['finalYrIdx']):
        for sex in GBRange(GB_Male, GB_Female):
            for data in [DP_Data, DP_Default]:
                result[data][sex][DP_AUnder1][t] = DPGlobalData['Covid19AgeDist'][DP_A0_4] * 0.67
                result[data][sex][DP_A1_4][t] = DPGlobalData['Covid19AgeDist'][DP_A0_4] * 0.33

                for age in GBRange(DP_A5_9, DP_MAX_AGE):
                        result[data][sex][age][t] = DPGlobalData['Covid19AgeDist'][age]

    return result

def UALowHighRatios_Init():
    return []

def AIDSDeathsFYearUA_Init():
    return 0

def UAHasRan_Init():
    return False

def UAInfo_Init():
    return {
        'ProcessDate'   : '',
        'ProcessDateDelphi' : 0.0,      #dies with delphi desktop
        'ProcessTime'   : '', 
        'NumIterations' : DPUA_Default_Num_Iterations,
        'AgDataYear'    : 2022,
    }